from copy import copy, deepcopy
from functools import lru_cache
from pathlib import Path
import warnings

from monty.json import MSONable
import numpy as np
from pymatgen.io.vasp.sets import DictSet
from pymatgen.io.vasp.inputs import Incar as pmgIncar
from pymatgen.io.vasp.inputs import Kpoints as pmgKpoints
from pymatgen.io.vasp.inputs import Poscar as pmgPoscar

from lightshow import _get_POTCAR_DIRECTORY_from_environ
from lightshow.parameters._base import _BaseParameters
from lightshow.common.kpoints import GenericEstimatorKpoints
from lightshow.common.nbands import UnitCellVolumeEstimate


VASP_INCAR_DEFAULT_NEUTRAL_POTENTIAL = {
    "ALGO": "Normal",
    "EDIFF": 1e-05,
    "IBRION": 2,
    "ISIF": 2,
    "ISMEAR": 0,
    "ISPIN": 2,
    "ISYM": 2,
    "KPAR": 1,
    "LAECHG": True,
    "LCHARG": True,
    "LORBIT": 11,
    "LREAL": "Auto",
    "LWAVE": False,
    "NCORE": 1,
    "NELM": 200,
    "NELMIN": 6,
    "NSIM": 16,
    "NSW": 0,
    "PREC": "Accurate",
    "SIGMA": 0.05,
    "SYMPREC": 1e-05,
}

VASP_INCAR_DEFAULT_COREHOLE_POTENTIAL = {
    "ALGO": "Normal",
    "CH_LSPEC": True,
    "CH_NEDOS": 40000,
    "CH_SIGMA": 0.05,
    "CLL": 0,
    "CLN": 1,
    "CLNT": 1,
    "CLZ": 1.0,
    "EDIFF": 1e-05,
    "EDIFFG": -0.01,
    "ICORELEVEL": 2,
    "ISMEAR": 0,
    "ISPIN": 2,
    "KPAR": 1,
    "LAECHG": True,
    "LCHARG": False,
    "LREAL": "Auto",
    "LWAVE": False,
    "NCORE": 1,
    "NELM": 200,
    "NSIM": 16,
    "PREC": "Accurate",
    "SIGMA": 0.05,
    "SYMPREC": 1e-05,
}

VASP_VALENCE_MAPPING = {
    "Ti": 12,
    "Zn": 20,
    "O": 6,
    "H": 1,
    "He": 2,
    "Li": 3,
    "Be": 4,
    "B": 3,
    "C": 4,
    "N": 5,
    "F": 7,
    "Ne": 8,
    "Na": 9,
    "Mg": 10,
    "Al": 3,
    "Si": 4,
    "P": 5,
    "S": 6,
    "Cl": 7,
    "Ar": 8,
    "K": 9,
    "Ca": 10,
    "Sc": 11,
    "V": 13,
    "Cr": 14,
    "Mn": 15,
    "Fe": 16,
    "Co": 17,
    "Ni": 18,
    "Cu": 19,
    "Ga": 21,
    "Ge": 22,
    "As": 23,
    "Se": 24,
    "Br": 25,
    "Kr": 8,
    "Rb": 9,
    "Sr": 10,
    "Y": 11,
    "Zr": 12,
    "Nb": 13,
    "Mo": 14,
    "Tc": 15,
    "Ru": 16,
    "Rh": 17,
    "Pd": 18,
    "Ag": 19,
    "Cd": 20,
    "In": 13,
    "Sn": 14,
    "Sb": 15,
    "Te": 6,
    "I": 7,
    "Xe": 8,
    "Cs": 9,
    "Ba": 10,
    "La": 11,
    "Ce": 12,
    "Hf": 12,
    "Ta": 13,
    "W": 14,
    "Re": 15,
    "Os": 16,
    "Ir": 17,
    "Pt": 18,
    "Au": 19,
    "Hg": 20,
    "Tl": 15,
    "Pb": 16,
    "Bi": 17,
    "Po": 18,
    "At": 17,
    "Rn": 18,
}

VASP_POTCAR_DEFAULT_ELEMENT_MAPPING = {
    "Ti": "Ti_sv_GW",
    "Zn": "Zn_sv_GW",
    "O": "O_GW",
    "H": "H_GW",
    "He": "He_GW",
    "Li": "Li_sv_GW",
    "Be": "Be_sv_GW",
    "B": "B_GW",
    "C": "C_GW",
    "N": "N_GW",
    "F": "F_GW",
    "Ne": "Ne_GW",
    "Na": "Na_sv_GW",
    "Mg": "Mg_sv_GW",
    "Al": "Al_GW",
    "Si": "Si_GW",
    "P": "P_GW",
    "S": "S_GW",
    "Cl": "Cl_GW",
    "Ar": "Ar_GW",
    "K": "K_sv_GW",
    "Ca": "Ca_sv_GW",
    "Sc": "Sc_sv_GW",
    "V": "V_sv_GW",
    "Cr": "Cr_sv_GW",
    "Mn": "Mn_sv_GW",
    "Fe": "Fe_sv_GW",
    "Co": "Co_sv_GW",
    "Ni": "Ni_sv_GW",
    "Cu": "Cu_sv_GW",
    "Ga": "Ga_d_GW",
    "Ge": "Ge_d_GW",
    "As": "As_GW",
    "Se": "Se_GW",
    "Br": "Br_GW",
    "Kr": "Kr_GW",
    "Rb": "Rb_sv_GW",
    "Sr": "Sr_sv_GW",
    "Y": "Y_sv_GW",
    "Zr": "Zr_sv_GW",
    "Nb": "Nb_sv_GW",
    "Mo": "Mo_sv_GW",
    "Tc": "Tc_sv_GW",
    "Ru": "Ru_sv_GW",
    "Rh": "Rh_sv_GW",
    "Pd": "Pd_sv_GW",
    "Ag": "Ag_sv_GW",
    "Cd": "Cd_sv_GW",
    "In": "In_d_GW",
    "Sn": "Sn_d_GW",
    "Sb": "Sb_d_GW",
    "Te": "Te_GW",
    "I": "I_GW",
    "Xe": "Xe_GW",
    "Cs": "Cs_sv_GW",
    "Ba": "Ba_sv_GW",
    "La": "La_GW",
    "Ce": "Ce_GW",
    "Hf": "Hf_sv_GW",
    "Ta": "Ta_sv_GW",
    "W": "W_sv_GW",
    "Re": "Re_sv_GW",
    "Os": "Os_sv_GW",
    "Ir": "Ir_sv_GW",
    "Pt": "Pt_sv_GW",
    "Au": "Au_sv_GW",
    "Hg": "Hg_sv_GW",
    "Tl": "Tl_d_GW",
    "Pb": "Pb_d_GW",
    "Bi": "Bi_d_GW",
    "Po": "Po_d_GW",
    "At": "At_d_GW",
    "Rn": "Rn_d_GW",
}


class Incar(pmgIncar):
    """Inherits the Pymatgen Incar class. Allows for the standard manipulation
    of the INCAR file as a dictionary but with a few extra methods."""

    def adj_mag(
        self,
        struct,
        config_dict={
            "INCAR": {
                "LASPH": True,
                "MAGMOM": {
                    "Ce": 5,
                    "Ce3+": 1,
                    "Co": 0.6,
                    "Co3+": 0.6,
                    "Co4+": 1,
                    "Cr": 5,
                    "Dy3+": 5,
                    "Er3+": 3,
                    "Eu": 10,
                    "Eu2+": 7,
                    "Eu3+": 6,
                    "Fe": 5,
                    "Gd3+": 7,
                    "Ho3+": 4,
                    "La3+": 0.6,
                    "Lu3+": 0.6,
                    "Mn": 5,
                    "Mn3+": 4,
                    "Mn4+": 3,
                    "Mo": 5,
                    "Nd3+": 3,
                    "Ni": 5,
                    "Pm3+": 4,
                    "Pr3+": 2,
                    "Sm3+": 5,
                    "Tb3+": 6,
                    "Tm3+": 2,
                    "V": 5,
                    "W": 5,
                    "Yb3+": 1,
                },
            },
            "KPOINTS": {"reciprocal_density": 100},
        },
    ):
        """Adjusts the INCAR to use the default magmom and +U from Pymatgen.

        Parameters
        ----------
        struct : pymatgen.core.structure.Structure
            The structure data for which the INCAR will be built.
        config_dict : dict, optional
            The default MAGMOM settings from the Materials Project.
        """

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            incar_tmp = DictSet(struct, config_dict).incar

        # Using spin unpolarized calculations for nm
        magmom_set = set(incar_tmp["MAGMOM"])
        if len(magmom_set) == 1 and 0 in magmom_set:
            incar_tmp["ISPIN"] = 1
            incar_tmp.pop("MAGMOM")

        # Assign the tags & values of magmoms and U to the INCAR
        for k, v in incar_tmp.items():
            self[k] = v

    def adj_u(
        self,
        elements,
        config_dict={
            "Co": 3.32,
            "Cr": 3.7,
            "Fe": 5.3,
            "Mn": 3.9,
            "Mo": 4.38,
            "Ni": 6.2,
            "V": 3.25,
            "W": 6.2,
        },
    ):
        """Updates the INCAR using the Materials Project default values for the
        Hubburd U values.

        Parameters
        ----------
        elements : list
            A list of str, containing the elements in the same order as the
            element line in the POSCAR.
        config_dict : dict, optional
            Default Materials Project U values for the transition metal
            elements.
        """

        tm = list(config_dict.keys())
        o_f = ["O", "F"]  # TODO: Fanchen, why just O and F?
        len_tm = len([ele for ele in tm if ele in elements])
        len_of = len([ele for ele in o_f if ele in elements])
        if len_tm > 0 and len_of > 0:
            self["LDAU"] = True
            self["LDAUJ"] = [0 for i in range(len(elements))]
            self["LDAUL"] = [2 for i in range(len(elements))]
            self["LDAUPRINT"] = 1
            self["LDAUTYPE"] = 2
            ldauu = []
            for ele in elements:
                if ele in config_dict:
                    ldauu.append(config_dict[ele])
                else:
                    ldauu.append(0)
            self["LDAUU"] = ldauu

    @classmethod
    def from_default(cls, neutral=True):
        """A simple classmethod for loading the default specs.

        Parameters
        ----------
        neutral : bool, optional
            If True, returns the default for a neutral potential electronic
            structure relaxation. Otherwise, assumes a corehole potential.

        Returns
        -------
        Incar
        """

        if neutral:
            klass = cls(VASP_INCAR_DEFAULT_NEUTRAL_POTENTIAL)
        else:
            klass = cls(VASP_INCAR_DEFAULT_COREHOLE_POTENTIAL)
        return klass

    def check_params(self):
        """Raises warnings for erroneous INCAR tags/parameters (see the
        `pymatgen.io.vasp.inputs documentation <https://pymatgen.org/
        pymatgen.io.vasp.inputs.html>`_ for precisely what this does). In
        addition, will actually raise a ValueError if "NBANDS" is not present
        in the INCAR.

        Raises
        ------
        ValueError
            If "NBANDS" is not in the INCAR ``keys``.
        """

        if "NBANDS" not in self.keys():
            raise ValueError("NBANDS not present in INCAR parameters")
        super().check_params()

    def write(self, target_directory):
        """Writes the INCAR file to the target directory.

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the INCAR file.
        """

        self.write_file(Path(target_directory) / Path("INCAR"))


class Kpoints(pmgKpoints):
    """The Pymatgen Kpoints object but with a redefined write method."""

    def write(self, target_directory):
        """Writes the KPOINTS file to the target directory.

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the KPOINTS file.
        """

        self.write_file(Path(target_directory) / Path("KPOINTS"))


class PotcarConstructor(MSONable):
    """A helper class used to create the POTCAR input files for VASP
    calculations. The default choices for the types of potentials chosen is
    detailed in :class:`.PotcarConstructor.VASP_POTCAR_DEFAULT_ELEMENT_MAPPING`
    and is chosen from the recommended potentials for GW/RPA calculations found
    `here <https://www.vasp.at/wiki/index.php/
    Available_PAW_potentials#Recommended_potentials_for_GW.
    2FRPA_calculations>`_.

    .. warning::

        The potential files fall under the VASP license. If you do not have a
        VASP license, you cannot use VASP potential files. Please take a look
        at the VASP `website <https://www.vasp.at/>`_ for more details.

    Parameters
    ----------
    root : str
        The location of the potential file-containing directories, e.g.,
        ``"Ti_sv_GW"``.
    element_mapping : dict
        Overrides specific defaults in the
        `VASP_POTCAR_DEFAULT_ELEMENT_MAPPING`. Note that only provided values
        will be overridden, and others will fallback to the defaults (unless
        ``override_default_element_mapping`` is True).
    override_default_element_mapping : bool
        If True, the ``element_mapping`` provided will override the default
        completely. If False, then only keys that are provided will be used
        to update the default mapping.
    """

    def __init__(
        self,
        root=None,
        element_mapping=dict(),
        override_default_element_mapping=False,
    ):
        if root is None:
            self._root = None
        else:
            self._root = Path(root).resolve()
            if not Path(self._root).exists():
                warnings.warn(
                    f"POTCAR directory {self._root} does not exist. POTCAR "
                    "files will not be written"
                )
                self._root = None

        self._override_default_element_mapping = (
            override_default_element_mapping
        )

        if self._override_default_element_mapping:
            self._element_mapping = element_mapping
        else:
            self._element_mapping = copy(VASP_POTCAR_DEFAULT_ELEMENT_MAPPING)
            self._element_mapping.update(element_mapping)

    def check_POTCAR_exists(self, list_of_elements):
        """Iterates through a list of element symbols and checks to see that a
        POTCAR file exists for each one of them.

        Parameters
        ----------
        list_of_elements : list
            A list of str, e.g., ``["Ti", "O"]``.

        Returns
        -------
        list
            A list of bool, where each entry corresponds to the true/false
            value of the POTCAR for that element existing.
        """

        el = []
        for element in list_of_elements:
            try:
                element_dir = self._element_mapping[element]
                path = Path(self._root) / Path(element_dir) / Path("POTCAR")
                el.append(path.exists())
            except KeyError:
                el.append(False)
        return el

    def check_POSCAR_valid(self, poscar):
        """Checks the poscar against the potcar data to ensure that all elements
        have a valid potential.

        Parameters
        ----------
        poscar : Poscar

        Returns
        -------
        str
            The elements that do not exist, else None if no error.
        """

        s = poscar.site_symbols
        elements_exist = self.check_POTCAR_exists(s)

        if not all(elements_exist):
            dont_exist = ",".join(
                [
                    xx
                    for xx, tf in zip(poscar.site_symbols, elements_exist)
                    if not tf
                ]
            )
            return dont_exist

        return None

    @lru_cache(256)
    def _get_element_lines(self, element):

        # Get the element directory using the element as a key
        assert isinstance(element, str)
        element_dir = self._element_mapping[element]

        # Construct the path
        path = Path(self._root) / Path(element_dir) / Path("POTCAR")

        # Load in the file
        with open(path, "r") as f:
            lines = f.readlines()

        return lines

    @lru_cache(256)
    def get_n_valence_electrons(self, element):
        """For a given element, reads the total number of valence electrons
        from the second line of the potential file.

        Parameters
        ----------
        element : str
            The element, e.g., ``"Ti"``.

        Returns
        -------
        int
            The number of valence electrons for that atom as specified by the
            potential file.
        """

        element_dir = self._element_mapping[element]

        # Construct the path
        path = Path(self._root) / Path(element_dir) / Path("POTCAR")

        # Load in the file
        with open(path, "r") as f:
            _ = f.readline()
            nelec = f.readline()

        return int(float(nelec.strip()))

    def get_total_valence_electrons(self, structure):
        """Gets the total number of valence electrons in a structure by
        counting the number of valence electrons per element.

        Parameters
        ----------
        structure : pymatgen.core.structure.Structure

        Returns
        -------
        int
            The total valence electrons in that structure. Note it counts the
            total valence in the full structure provided, so supercells will
            have more total valence electrons than primitive cells.
        """

        elements = [str(s.specie) for s in structure]
        return int(sum([self.get_n_valence_electrons(ii) for ii in elements]))

    def write(self, target_directory, elements):
        """Writes the POTCAR file to the path provided by stacking the
        potentials for the provided elements.

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the POTCAR file.
        elements : list
            A list of str: the elements to use in the POTCAR file, in the
            correct order.
        """

        if self._root is None:
            return

        path = Path(target_directory) / Path("POTCAR")

        lines = []
        for element in elements:
            lines.extend(self._get_element_lines(element))

        with open(path, "w") as f:
            for line in lines:
                f.write(line)


class Poscar(pmgPoscar):
    """Custom Poscar file, lightweight wrapper for the
    :class:`pymatgen.io.vasp.inputs.Poscar` class.

    .. warning::

        This :class:`.Poscar` class is only tested with VASP >6.2. It is likely
        compatible with earlier versions, but the user should proceed with
        caution in those cases.
    """

    def __len__(self):
        return len(self.structure)

    def write_single_site(
        self,
        target_directory,
        site_index=None,
        check_atom_type=None,
        significant_figures=8,
    ):
        """Writes the VASP POSCAR file for the specified ``site_index``.

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the POSCAR file.
        site_index : int, optional
            In corehole calculations, we must move the site of interest to the
            first entry in the POSCAR file, and ensure that the element lines
            in said POSCAR are changed accordingly. For example,
            ``Ti O -> Ti Ti O`` and ``6 10 -> 1 5 10``, are changed in order to
            properly reflect the new ordering of atoms. This is to simplify how
            the POTCAR file is saved, and indicates for VASP to use a different
            potential for the atom with the corehole on it. Default is None
            (indicating no changes to be made, this is useful for neutral
            potential calculations).
        check_atom_type : str, optional
            If not None, asserts that the found atom type in the POSCAR lines
            is the same as the provided atom type.
        significant_figures : int, optional
            The number of significant digits to write.

        Returns
        -------
        list
            A list of str: the ordering of the elements for the POTCAR file.
        """

        path = Path(target_directory) / Path("POSCAR")

        # This might be specific to certain versions of VASP
        STRUCTURE_START_INDEX = 8

        # In the case of e.g. the neutral potential calculation, the base class
        # method write_file can be used
        if site_index is None:
            self.write_file(path, significant_figures=significant_figures)
            return self.site_symbols

        # Get the default lines for the POSCAR file
        lines = self.get_string(significant_figures=significant_figures).split(
            "\n"
        )

        # We need to make some modifications depending on the type of
        # calculation this is. Lines indexed by 5 and 6 are the ones that
        # need to be changed, and lines 8 onward contain the structure data
        # to be swapped.
        assert isinstance(site_index, int)
        assert site_index >= 0
        site_index += STRUCTURE_START_INDEX

        # Get the atom type of the site provided
        atom_type = lines[site_index].split()[-1]

        # Assert if provided
        if check_atom_type is not None:
            assert atom_type == check_atom_type

        # Get the location of the atom type
        atom_type_loc = np.where(np.array(self.site_symbols) == atom_type)[
            0
        ].item()
        new_line_5 = [atom_type, *self.site_symbols]
        line_6_asint = [int(xx) for xx in self.natoms]
        line_6_asint[atom_type_loc] -= 1
        new_line_6_asint = [1, *line_6_asint]
        new_line_6 = [str(xx) for xx in new_line_6_asint]

        # Use the new lines 5 and 6
        lines[5] = " ".join(new_line_5)
        lines[6] = " ".join(new_line_6)

        # Execute the swap. First, get the line of interest
        move_to_front = lines[site_index]

        # Delete that entry from the list
        del lines[site_index]

        # And insert back at the top
        lines.insert(STRUCTURE_START_INDEX, move_to_front)

        # And save
        with open(path, "w") as f:
            for line in lines:
                f.write(f"{line}\n")

        return new_line_5


class VASPParameters(MSONable, _BaseParameters):
    """A one-stop-shop for modifying the VASP parameters for some calculation.
    Due to the relative complexity of setting up a VASP calculation (as opposed
    to e.g. FEFF), the :class:`.VASPParameters` object...

    Parameters
    ----------
    incar : dict or :class:`.Incar`
        The INCAR VASP parameter information. If of type dict, then an
        :class:`.Incar` class will be instantiated from it. Can also be of type
        :class:`.Incar` directly. The number of bands does not have to be
        directly specified, as it can be estimated based on the structural
        information as long as ``nbands_estimator`` is not ``None``.
    edge : str
        The edge of the spectroscopy calculation.
    potcar_directory : os.PathLike, optional
        The location in which the potential files are stored. These files
        should be stored in a precise directory format, specifically: the
        first level in should contain potential files of the form
        ``Pb_d_GW``, ``Pb_sv_GW``, ``Pd``, etc. For some element, the specific
        choice is determined first by the ``potcar_element_mapping`` and then
        by the defaults in the :class:`.PotcarConstructor` class. If None,
        checks the environment for ``VASP_POTCAR_DIRECTORY``.
    kpoints : lightshow.common.kpoints._BaseKpointsMethod
        The method for constructing he kpoints file from the structure. Should
        be a class with a ``__call__`` method defined. This method should take
        the structure as input and return a tuple corresponding to the kpoints
        density along each axis.
    nbands : lightshow.common.nbands._BaseNbandsMethod
        The method for determining the number of valence bands from the
        structure. Should be a class with a ``__call__`` method defined. This
        method should take the structure as input and return an integer: the
        number of valence bands to use in the calculation.
    potcar_element_mapping : dict
        A custom element mapping which will override, element-by-element,
        defaults in the :class:`.PotcarConstructor` class.
    max_bands : int
        The maximum number of bands to use in a calculation. This is a failsafe
        in case the ``nbands`` method produces a result with more bands than
        can be feasibly run on your computer.
    force_spin_unpolarized : bool
        TODO
    name : str
        The name of the calculation. Should likely always be ``"VASP"``.
    """

    def __init__(
        self,
        incar,
        edge=None,
        potcar_directory=None,
        kpoints=GenericEstimatorKpoints(cutoff=32.0, max_radii=50.0),
        nbands=UnitCellVolumeEstimate(e_range=40.0),
        potcar_element_mapping=dict(),
        max_bands=-1,
        force_spin_unpolarized=False,
        name="VASP",
    ):

        if "NBANDS" in incar.keys():
            if incar["NBANDS"] is None and nbands is None:
                raise ValueError(
                    "One of incar['NBANDS'] or the nbands argument must be "
                    "provided"
                )

            if incar["NBANDS"] is not None and nbands is not None:
                warnings.warn(
                    "Constant incar['NBANDS'] will override provided nbands "
                    "method."
                )

        if edge is not None:
            warnings.warn(
                f"edge={edge} was specified. Currently this method is not "
                "implemented, and the edge should be specified by setting "
                "the CLN and CLL parameters in the incar. See "
                "https://www.vasp.at/wiki/index.php/XANES_in_Diamond#INCAR"
            )
        self._edge = edge

        # Load the INCAR information
        if isinstance(incar, Incar):
            self._incar = incar
        elif isinstance(incar, dict):
            self._incar = Incar(incar)
        else:
            raise ValueError(
                f"Unknown incar type: {type(incar)}. Should be of type "
                "Incar or dict"
            )

        # Method for generating the KPOINTS file
        self._kpoints = kpoints

        # POTCAR information
        if potcar_directory is None:
            potcar_directory = _get_POTCAR_DIRECTORY_from_environ()
        if potcar_directory is None:
            warnings.warn(
                "potcar_directory not provided, and VASP_POTCAR_DIRECTORY "
                "not in the current environment variables. POTCAR files will "
                "not be written."
            )

        self._potcar_directory = potcar_directory
        self._potcar_element_mapping = potcar_element_mapping
        self._potcar_constructor = PotcarConstructor(
            potcar_directory, potcar_element_mapping
        )

        # Estimator method for the number of bands
        self._nbands = nbands

        # Other parameters
        if max_bands == -1:
            self._max_bands = int(1e16)
        else:
            self._max_bands = max_bands
        self._force_spin_unpolarized = force_spin_unpolarized
        self._name = name

    @staticmethod
    def _vbands(structure):
        """Get the number of the valence bands using the
        ``VASP_VALENCE_MAPPING``.

        Parameters
        ----------
        pymatgen.core.structure.Structure

        Returns
        -------
        int

        Raises
        ------
        KeyError
            If a particular species string is not found in the
            ``VASP_VALENCE_MAPPING``.
        """

        return sum(
            [VASP_VALENCE_MAPPING[atom.species_string] for atom in structure]
        )

    def write(self, target_directory, **kwargs):
        """Summary

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the VASP input files.
        **kwargs
            Must contain the ``structure_sc`` key (the
            :class:`pymatgen.core.structure.Structure` of interest) and the
            ``sites`` key (a list of int, where each int corresponds to the
            site index of the site to write).

        Returns
        -------
        dict
            A dictionary containing the status and errors key.

        Raises
        ------
        ValueError
            If the symmetrically inequivalent sites provided is None but the
            CH_LSPEC flag is True in the INCAR.
        """

        errors = dict()

        incar = deepcopy(self._incar)
        sites = kwargs.get("sites")
        index_mapping = kwargs.get("index_mapping")
        ch_lspec = incar.get("CH_LSPEC", False)
        if sites is None and ch_lspec:
            raise ValueError(
                "Corehole calculation has been selected via the presence of "
                "CH_LSPEC in the INCAR, but no sites were provided."
            )

        structure = deepcopy(kwargs["structure_sc"])

        # Set the number of bands if the number of bands is not explicitly
        # provided
        if incar.get("NBANDS") is None:
            cb = self._nbands(structure)
            vb = self._vbands(structure)
            nb = cb + vb
            if nb > self._max_bands:
                errors["n_bands"] = nb
            incar["NBANDS"] = nb

        poscar = Poscar(structure)
        if self._potcar_constructor._root is not None:
            potcar_check = self._potcar_constructor.check_POSCAR_valid(poscar)
            if potcar_check is not None:
                errors["potentials_do_not_exist"] = potcar_check

        if self._force_spin_unpolarized:
            incar["ISPIN"] = 1
            try:
                incar.pop("MAGMOM")
            except KeyError:
                pass
        else:
            incar.adj_mag(structure)
        incar.check_params()

        if len(errors) > 0:
            return {"pass": False, "errors": errors}

        k = self._kpoints(structure)
        kpoints = Kpoints(kpts=(k,))
        species = [
            structure[index_mapping[site]].specie.symbol for site in sites
        ]

        paths = []

        # Green light for corehole calculations
        if ch_lspec:
            for site, specie in zip(sites, species):
                path = target_directory / Path(f"{site:03}_{specie}")
                path.mkdir(exist_ok=True, parents=True)
                elements = poscar.write_single_site(
                    path,
                    site_index=index_mapping[site],
                    check_atom_type=specie,
                )
                self._potcar_constructor.write(path, elements)
                kpoints.write(path)
                incar.adj_u(elements)
                incar.write(path)
                paths.append(path)

        # Else assume this is just a standard neutral potential calculation
        else:
            path = target_directory / Path("SCF")
            path.mkdir(exist_ok=True, parents=True)
            elements = poscar.write_single_site(
                path, site_index=None, check_atom_type=None
            )
            self._potcar_constructor.write(path, elements)
            kpoints.write(path)
            incar.adj_u(elements)
            incar.write(path)
            paths.append(path)

        return {"pass": True, "errors": dict(), "paths": paths}



from copy import deepcopy
from pathlib import Path
import pytest

from pymatgen.core.structure import Structure

from lightshow import Database

STRUCTURE_FILES_PATH = (
    Path.cwd() / Path("lightshow") / Path("_tests") / Path("structure_files")
)


@pytest.fixture
def test_structure_names():
    return [
        "mp-390",
        "mvc-11115",
        "mp-1215",
        "mp-1840",
        "mp-2657",
        "mp-2664",
        "mp-430",
        "mp-458",
        "mp-10734",
    ]


@pytest.fixture
def database_from_file():
    dat = Database.from_files(
        STRUCTURE_FILES_PATH, filename="POSCAR", cleanup_paths=True
    )
    return deepcopy(dat)


@pytest.fixture
def dummy_potcar_file_directory():
    return str(Path(__file__).parent.resolve() / Path("dummy_potcar_files"))


@pytest.fixture
def dummy_psp_file_directory():
    return str(Path(__file__).parent.resolve() / Path("dummy_psp_files"))


@pytest.fixture
def dummy_chpsp_file_directory():
    return str(Path(__file__).parent.resolve() / Path("dummy_chpsp_files"))


@pytest.fixture
def mp_Structure_mp390():
    return Structure.from_file(
        STRUCTURE_FILES_PATH / Path("mp-390") / Path("POSCAR")
    )



import numpy as np
from pymatgen.io.ase import AseAtomsAdaptor as ase
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


def make_supercell(prim, cutoff=9.0):
    """Used to generate supercell with desired lattice vector: default cutoff
    9 Angstrom, which is am empirical value according to our test. It will
    choose either to use primitive unit cell or to use  conventional unit cell
    depending on which has less atoms.

    Parameters
    ----------
    prim : pymatgen.core.structure.Structure
        The primitive structure.
    cutoff : float, optional
        The supercell cutoff.

    Returns
    -------
    pymatgen.core.structure.Structure
    """

    # If no MAGMOM data, setup for spin polarized calculations
    if "magmom" not in prim.site_properties:
        magmom = [1 for i in range(len(prim))]
        prim.add_site_property("magmom", magmom)
    prim_magmom = prim.site_properties["magmom"]

    latt_prim = np.array(prim.lattice.abc)

    shape_prim = np.ceil(cutoff / latt_prim).astype(int)

    prim_super = ase.get_atoms(prim, magmoms=prim_magmom) * shape_prim

    magmom = prim_super.get_initial_magnetic_moments()

    structure = ase.get_structure(prim_super)

    for idx, site in enumerate(structure):
        site.properties["magmom"] = magmom[idx]

    return structure


def get_inequivalent_site_info(structure):
    """Gets the symmetrically inequivalent sites as found by the
    SpacegroupAnalyzer class from Pymatgen.

    Parameters
    ----------
    structure : pymatgen.core.structure.Structure
        The Pymatgen structure of interest.

    Returns
    -------
    dict
        A dictionary containing three lists, one of the inequivalent sites, one
        for the atom types they correspond to and the last for the multiplicity.
    """

    # Get the symmetrically inequivalent indexes
    inequivalent_sites = (
        SpacegroupAnalyzer(structure)
        .get_symmetrized_structure()
        .equivalent_indices
    )

    # Equivalent indexes must all share the same atom type
    multiplicities = [len(xx) for xx in inequivalent_sites]
    inequivalent_sites = [xx[0] for xx in inequivalent_sites]
    species = [str(structure[xx].specie) for xx in inequivalent_sites]

    return {
        "sites": inequivalent_sites,
        "species": species,
        "multiplicities": multiplicities,
    }



from abc import ABC, abstractmethod

from monty.json import MSONable
import numpy as np


class _BaseKpointsMethod(ABC):
    @abstractmethod
    def __call__(self, structure):
        """Base abstraction for the various ways to get the kpoints for any
        simulation. This method should take the structure as input, and return
        a tuple containing the desired kmesh.

        Parameters
        ----------
        structure : pymatgen.core.structure.Structure

        Returns
        -------
        tuple
        """

        ...


class Constant(MSONable, _BaseKpointsMethod):
    def __init__(self, kmesh):
        self._kmesh = kmesh

    def __call__(self, structure):
        return tuple(self._kmesh)


class GenericEstimatorKpoints(MSONable, _BaseKpointsMethod):
    """For a kmesh sampling, e.g. [m, n, p], of a crystal "cell" is equivalent
    to generating a supercell with [m, n, p] the crystal cell. The
    corresponding radius is the largest radius of the sphere that can fit into
    this supercell. The radius can also be regarded as the inverse kmesh
    density. Along each reciprocal lattice, the kmesh densities are not
    guaranteed to be the same. The smallest kmesh density is chosen. This
    function uses the effective radius as a controlling factor to determining
    the kemsh.

    Parameters
    ----------
    cutoff : float, optional
        Cutoff radius for constructing the kmesh. It will loop a look-up
        table for the effective radius (controlled by max_radii). The kmesh
        with radius right above the cutoff will be chosen. Default is 32
        Angstroms (60 Bohr).
    max_radii : float, optional
        Maximum radius used for constructing the lookup table.
    """

    def __init__(self, cutoff=32.0, max_radii=50.0):
        self._cutoff = cutoff
        self._max_radii = max_radii

    def __call__(self, structure):
        klist = dict()
        rlatt = np.array(structure.lattice.reciprocal_lattice.abc)

        # TODO: why 10, 0.2?
        for xx in np.arange(0, 10, 0.2):
            div = np.floor(xx * rlatt) + 1
            divlatt = 2.0 * np.pi / rlatt * div

            radi = min(divlatt)

            if radi > self._max_radii:
                break
            else:
                div = tuple(div.astype(int))
                if div not in klist:
                    klist[div] = radi

        for key, value in klist.items():
            if value > self._cutoff:
                k = key
                break

        return k



from copy import copy
from pathlib import Path
from warnings import warn

from monty.json import MSONable
from pymatgen.io.feff.sets import MPXANESSet, MPEXAFSSet, FEFFDictSet

from lightshow.parameters._base import _BaseParameters


FEFF_DEFAULT_CARDS = {
    "S02": "0",
    "COREHOLE": "RPA",
    "CONTROL": "1 1 1 1 1 1",
    "XANES": "4 0.04 0.1",
    "SCF": "7.0 0 100 0.2 3",
    "FMS": "9.0 0",
    "EXCHANGE": "0 0.0 0.0 2",
    "RPATH": "-1",
}


class FEFFParameters(MSONable, _BaseParameters):
    """A one-stop-shop for all the different ways to modify input parameters
    for a FEFF calculation. This class is a lightweight wrapper for the
    ``FEFFDictSet``, containing some extra information and methods used for
    writing the appropriate input files. Like most Pymatgen objects, the
    class is serializable via e.g. ``feff_parameters.as_dict()``.

    Parameters
    ----------
    cards : dict
        A dictionary of of the cards to be used in the FEFF calculations.
        For example, for XANES, one might wish to use something like

        .. code-block:: python

            cards = {
                "S02": "0",
                "COREHOLE": "RPA",
                "CONTROL": "1 1 1 1 1 1",
                "XANES": "4 0.04 0.1",
                "SCF": "7.0 0 100 0.2 3",
                "FMS": "9.0 0",
                "EXCHANGE": "0 0.0 0.0 2",
                "RPATH": "-1"
            }

        And for EXAFS,

        .. code-block:: python

            cards = {
                "S02": "0",
                "COREHOLE": "RPA",
                "CONTROL": "1 1 1 1 1 1",
                "EXAFS": "20",
                "RPATH": "10",
                "SCF": "7.0 0 100 0.2 3",
                "EXCHANGE": "0 0.0 0.0 0"
            }
    edge : str
        The XAS edge of the calculation.
    radius : float
        FEFF uses clusters for its calculations. The ``radius`` parameter
        determines how large to make the cluster. It is calculated from the
        absorbing atom center in units of Angstroms.
    spectrum : str
        The type of spectroscopy to be run. This is used to set the default
        cards through Pymatgen. These defaults can be overridden by setting
        the ``cards`` argument when instantiating the class. The type of
        calculation to run. Should likely be either ``"XANES"`` or ``"EXAFS"``.
    **feff_dict_set_kwargs
        Keyword arguments to pass directly to the FEFFDictSet object before
        writing the input files. For example, ``nkpts`` can be used to specify
        the number of k-points used in the Brillouin zone. It is only used if
        FEFF is run in reciprocal space mode.
    """

    def __init__(
        self,
        cards=FEFF_DEFAULT_CARDS,
        edge="K",
        radius=9.0,
        spectrum="XANES",
        name="FEFF",
        **feff_dict_set_kwargs,
    ):
        # Try to see if "edge" is in the provided card keys
        if "EDGE" in cards.keys():
            warn(f"Provided edge in cars will be overwritten by kwarg {edge}")
            cards.pop("EDGE")

        self._cards = cards
        self._radius = radius
        self._spectrum = spectrum
        self._name = name
        self._edge = edge
        self._feff_dict_set_kwargs = feff_dict_set_kwargs

        if self._edge == "L":
            warn(
                "Specified edge is 'L' and will be changed to L3 to be "
                "FEFF-compatible"
            )
            self._edge = "L3"
        elif "M" in self._edge:
            warn(
                "According to the FEFF9 documentaiton, M-edges are not well "
                "tested; proceed with caution!"
            )
        elif self._edge not in ["K", "L1", "L2", "L3"]:
            warn(
                f"Provided edge {self._edge} is not one of the standard "
                "choices K, L1, L2, L3 and it is unknown if FEFF supports it"
            )

    def get_FEFFDictSets(self, structure, absorbing_sites):
        """Constructs and returns a list of the
        :class:`pymatgen.io.feff.sets.FEFFDictSet` objects.

        Parameters
        ----------
        structure : pymatgen.core.structure.Structure
            The Pymatgen structure. Note that the ``absorbing_sites`` must
            correspond to the provided structure.
        absorbing_sites : list
            A list of int corresponding to absorbing sites.

        Returns
        -------
        list
            A list of :class:`pymatgen.io.feff.sets.FEFFDictSet` objects. One
            for each absorbing site.

        Raises
        ------
        ValueError
            If an invalid ``spectrum`` argument is provided.
        """

        if self._spectrum == "XANES":
            default_cards = copy(MPXANESSet.CONFIG)
        elif self._spectrum == "EXAFS":
            default_cards = copy(MPEXAFSSet.CONFIG)
        else:
            raise ValueError(f"Unknown spectrum type {self._spectrum}")

        default_cards.pop("EDGE")
        default_keys = default_cards.keys()
        keys_to_remove = list(set(default_keys) - set(self._cards.keys()))
        user_tag_settings = {"_del": keys_to_remove}
        user_tag_settings.update(self._cards)

        return [
            FEFFDictSet(
                site,
                structure,
                spectrum=self._spectrum,
                config_dict=default_cards,
                edge=self._edge,
                radius=self._radius,
                user_tag_settings=user_tag_settings,
                **self._feff_dict_set_kwargs,
            )
            for site in absorbing_sites
        ]

    def write(self, target_directory, **kwargs):
        """Writes the input files for the provided structure and sites. In the
        case of FEFF, if sites is None (usually indicating a global calculation
        such as a neutral potential electronic relaxation method in VASP), then
        write does nothing.

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the FEFF input files.
        **kwargs
            Must contain the ``structure`` key (the
            :class:`pymatgen.core.structure.Structure` of interest) and the
            ``sites`` key (a list of int, where each int corresponds to the
            site index of the site to write).

        Returns
        -------
        dict
            A dictionary containing the status and errors key. In the case of
            FEFF, there are no possible errors at this stage other than
            critical ones that would cause program termination, so the returned
            object is always
            ``{"pass": True, "errors": dict(), "paths": [...]}``.
        """

        structure = kwargs["structure_uc"]
        sites = kwargs["sites"]

        target_directory = Path(target_directory)
        target_directory.mkdir(exist_ok=True, parents=True)

        # Get the directory names
        species = [structure[site].specie.symbol for site in sites]
        names = [f"{site:03}_{specie}" for site, specie in zip(sites, species)]

        dict_sets = self.get_FEFFDictSets(structure, sites)

        paths = []
        for dict_set, name in zip(dict_sets, names):
            path = target_directory / Path(name)
            dict_set.write_input(path)
            paths.append(path)

        return {"pass": True, "errors": dict(), "paths": paths}



from copy import copy
from pathlib import Path
import numpy as np

from monty.json import MSONable
from pymatgen.core import Element

from lightshow.parameters._base import _BaseParameters
from lightshow.common.kpoints import GenericEstimatorKpoints
from lightshow.common.nbands import UnitCellVolumeEstimate


OCEAN_DEFAULT_CARDS = {
    "dft": "qe",
    "ecut": "-1",
    "opf.program": "hamann",
    "para_prefix": "mpirun -np 24",
}

Bohr = 0.5291772105638411


class OCEANParameters(MSONable, _BaseParameters):
    """A one-stop-shop for all the different ways to modify input parameters
    for an OCEAN calculation.

    Parameters
    ----------
    cards : dict
        A dictionary of of the cards to be control the parameters in the
        OCEAN calculations. The key of the dictionary corresponds to the
        parameters in OCEAN; the values are the correspongding values.
        In LightShow, minimum parameters to run ocean is provided in
        ``OCEAN_DEFAULT_CARDS``, which looks something like

        .. code-block:: python

            cards = {
                    "dft" : "qe",
                    "ecut" : "-1",
                    "opf.program" : "hamann",
                    "para_prefix" : "mpirun -np 24",
            }

        The detailed description of the OCEAN parameters can be find at
        its official website. If the user wants to change some parameters,
        they can just add a key-value pair to the cards.
    kpoints : lightshow.common.kpoints._BaseKpointsMethod
        The method for constructing he kpoints file from the structure. Should
        be a class with a ``__call__`` method defined. This method should take
        the structure as input and return a tuple corresponding to the kpoints
        density along each axis.
    nbands : lightshow.common.nbands._BaseNbandsMethod
        The method for determining the number of valence bands from the
        structure. Should be a class with a ``__call__`` method defined. This
        method should take the structure as input and return an integer: the
        number of valence bands to use in the calculation.
    defaultConvPerAtom: float, optional
        Enegy convergence threshold.
    bandgap: float, optional
        User can provide the band gap of the material they are interested in
        to determine the ``diemac``. If not provided and the strucutre is from
        pymatgen, the code will use the data (either band gap or diel) stored
        in the database to get ``diemac``; if not provided and the structure
        is from a file, the default value of ``diemac`` in the cards will be
        used. If ``diel`` is also not None, ``diemac`` is determined using
        ``diel``.
    diel: float, optional
        User can provide the diel of the material they are interested in
        to determine the ``diemac``. If not provided and the strucutre is from
        pymatgen, the code will use the the data (either band gap or diel)
        stored in the database to get ``diemac``; if not provided and the
        structure is from a file, the default value 10000 is used for  `diemac`.
    edge: str, optional
        The edge of the XAS to run in the calculation. The default is K.
    name : str
        The name of the calculation. Should likely always be ``"OCEAN"``.

    Raises
    ------
    ValueError
        If an invalid ``edge`` argument is provided.
    """

    @property
    def name(self):
        return self._name

    @property
    def cards(self):
        return self._cards

    def __init__(
        self,
        cards=OCEAN_DEFAULT_CARDS,
        kpoints=GenericEstimatorKpoints(cutoff=16.0, max_radii=50.0),
        nbands=UnitCellVolumeEstimate(e_range=30.0),
        bandgap=None,
        diel=None,
        defaultConvPerAtom=1e-10,
        edge="K",
        name="OCEAN",
    ):

        # Default cards for ocean
        self._cards = cards

        # User modified cards
        self._bandgap = bandgap
        self._diel = diel

        # Method for determining number of bands
        self._nbands = nbands

        # Method for determining the kmesh
        self._kpoints = kpoints

        self._defaultConvPerAtom = defaultConvPerAtom

        # Handle the edge parsing here.
        if edge not in self._edge_map.keys():
            raise ValueError(
                f"Provided edge {edge} is not a valid choice for OCEAN "
                "calculations. Edges should be chosen from the list: "
                f"{list(self._edge_map.keys())}"
            )
        self._edge = self._edge_map[edge]
        self._name = name

    @property
    def _edge_map(self):
        """Mapping between letter (indicating the edge) and number (input to
        OCEAN) defining the XAS edge."""

        return {
            "K": "1 0",
            "L": "2 1",
            "L1": "2 0",
            "L2": "2 1",
            "L3": "2 1",
            "L23": "2 1",
            "M": "3 2",
            "M1": "3 0",
            "M2": "3 1",
            "M3": "3 1",
            "M23": "3 1",
            "M4": "3 2",
            "M5": "3 2",
            "M45": "3 2",
            "N": "4 3",
            "N1": "4 0",
            "N2": "4 1",
            "N3": "4 1",
            "N23": "4 1",
            "N4": "4 2",
            "N5": "4 2",
            "N45": "4 2",
            "N6": "4 3",
            "N7": "4 3",
            "N67": "4 3",
        }

    @staticmethod
    def _oceanKptSampling(cell, kpt):
        """Method for getting the explicit ``screen.nkpt``. ``screen.nkpt``
        accepts a single value as input, which is similar to the idea of
        effecitve radius and is implemented in the latest version of OCEAN. If
        a sinlge value is provided, the code will convert it back to the
        original kmesh, something like ``3 3 2``.

        Parameters
        ----------
        cell : (3, 3) numpy array
            Lattice parameter in the matrix form
        kpt : float

        Returns
        -------
        str
            a string consisting the kmesh values along three dimentions
        """
        v = abs(np.dot(np.cross(cell[0], cell[1]), cell[2])) / Bohr
        b1 = 2 * np.pi * np.linalg.norm(np.cross(cell[1], cell[2])) / v
        b2 = 2 * np.pi * np.linalg.norm(np.cross(cell[0], cell[2])) / v
        b3 = 2 * np.pi * np.linalg.norm(np.cross(cell[0], cell[1])) / v

        k1 = int(-kpt * b1) + 1
        k2 = int(-kpt * b2) + 1
        k3 = int(-kpt * b3) + 1

        return "{:d} {:d} {:d}".format(k1, k2, k3)

    @staticmethod
    def _write_ocean_in(path, structure, input_data: dict):
        """Methods for writing the OCEAN input file, e.g. ``ocean.in``.
        The lattice structure and element type are handled here. The user
        provided input paramters will also be managed.

        Parameters
        ----------
        path : os.PathLike
            the target file where all the input paramters will be written.
        structure : pymatgen.core.structure.Structure
            The Pymatgen structure.
        input_data : dict
            A dictionary containing all the user defined parameters for
            OCEAN input file.

        """
        fd = open(path, "w")

        input_data_str = []
        for key in input_data:
            input_data_str.append(
                str(key) + " { " + str(input_data[key]) + " }\n"
            )

        fd.write("".join(input_data_str))
        species = sorted(set(structure.atomic_numbers))

        fd.write("znucl {{ {} }}\n".format(" ".join(str(Z) for Z in species)))
        fd.write("typat")
        fd.write("{\n")
        types = []
        for Z in structure.atomic_numbers:
            for n, Zs in enumerate(species):
                if Z == Zs:
                    types.append(n + 1)
        n_entries_int = 20  # integer entries per line
        for n, type in enumerate(types):
            fd.write(" %d" % (type))
            if n > 1 and ((n % n_entries_int) == 1):
                fd.write("\n")
        fd.write(" }\n")

        atomic_positions_str = []
        for atom in structure:
            atomic_positions_str.append(
                "{coords[0]:.10f} {coords[1]:.10f} {coords[2]:.10f}\n".format(
                    coords=[atom.a, atom.b, atom.c]
                )
            )

        fd.write("xred {\n")
        fd.write("".join(atomic_positions_str))
        fd.write("}\n")

        fd.write(
            "acell {{ {acell[0]} {acell[0]} {acell[0]} }} \n".format(
                acell=[1 / Bohr]
            )
        )

        fd.write(
            "rprim {{ {cell[0][0]:.14f} {cell[0][1]:.14f} {cell[0][2]:.14f}\n"
            "        {cell[1][0]:.14f} {cell[1][1]:.14f} {cell[1][2]:.14f}\n"
            "        {cell[2][0]:.14f} {cell[2][1]:.14f} {cell[2][2]:.14f}  }}\n"  # noqa
            "".format(cell=structure.lattice.matrix)
        )

    def write(self, target_directory, **kwargs):
        """Writes the input files for the provided structure and sites. Some
        user defined parameters will also be handled during this stage, such as
        ``diel`` or ``bandgap``.

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the FEFF input files.
        **kwargs
            Must contain the ``structure_uc`` key (the
            :class:`pymatgen.core.structure.Structure` of interest) and the
            ``sites`` key (a list of int, where each int corresponds to the
            site index of the site to write).

        Returns
        -------
        dict
            A dictionary containing the status and errors key. In the case of
            OCEAN, there are no possible errors at this stage other than
            critical ones that would cause program termination, so the returned
            object is always ``{"pass": True, "errors": dict()}``.
        """

        structure = kwargs["structure_uc"]
        sites = kwargs["sites"]

        target_directory = Path(target_directory)
        target_directory.mkdir(exist_ok=True, parents=True)

        # Obtain absorbing atom
        species = [structure[site].specie.symbol for site in sites]
        element_list = []
        for specie in species:
            element = Element(specie)
            if element in element_list:
                continue
            element_list.append(element)
            cards = copy(self._cards)

            cards["edges"] = ""
            for site in sites:
                s = structure[site].specie.symbol
                site += 1
                if s == specie:
                    cards["edges"] += f"{site} {self._edge}\n"
            cards["edges"] = cards["edges"].rstrip()

            # ocean use a negative value to represent the number of
            # conduction bands
            cards["nbands"] = -1 * self._nbands(structure)

            # Standardized methods for getting the kpoints
            kmesh = self._kpoints(structure)
            cards["ngkpt"] = f"{kmesh[0]} {kmesh[1]} {kmesh[2]}"

            # Determine the diemac
            if self._bandgap is not None or self._diel is not None:
                if self._diel is not None:
                    cards["diemac"] = self._diel
                elif self._bandgap is not None:
                    cards["diemac"] = np.exp(3.5 / self._bandgap)
            elif self._bandgap is None and self._diel is None:
                if "bandgap" in kwargs.keys() and "diel" in kwargs.keys():
                    bandgap = kwargs["bandgap"]
                    diel = kwargs["diel"]
                    if diel is not None:
                        if diel["poly_electronic"] is not None:
                            cards["diemac"] = diel["poly_electronic"]
                    elif bandgap is not None:
                        if bandgap > 0.000001:
                            cards["diemac"] = np.exp(3.5 / bandgap)
                        else:
                            cards["diemac"] = 1000000
            # Determine the SCF? convergence threshold
            cards["toldfe"] = self._defaultConvPerAtom * len(structure)
            # Change screen.nkpt -Int to triplet of ints
            if (
                "screen.nkpt" in cards
                and len(cards["screen.nkpt"].split()) == 1
            ):
                cards["screen.nkpt"] = self._oceanKptSampling(
                    structure.lattice.matrix, float(cards["screen.nkpt"])
                )

            # OCEAN will calculate every atom within the same specie
            path = target_directory / Path(specie)
            path.mkdir(exist_ok=True, parents=True)
            filepath_xas = path / "ocean.in"
            self._write_ocean_in(filepath_xas, structure, cards)

            # Deal with the dipole case only
            # notice I put the photonSymm in the folder, which is created by
            # John
            photons = list()
            photons.append({"dipole": [1, 0, 0, 1]})
            photons.append({"dipole": [0, 1, 0, 1]})
            photons.append({"dipole": [0, 0, 1, 1]})

            totalweight = 0
            for photon in photons:
                totalweight += photon["dipole"][3]

            photonCount = 0
            for photon in photons:
                photonCount += 1
                dir1 = photon["dipole"][0:3]
                dir2 = dir1
                weight = photon["dipole"][3] / totalweight
                mode = "dipole"

                with open(path / ("photon%d" % (photonCount)), "w") as f:
                    f.write(mode + "\n")
                    f.write(
                        "cartesian %f %f %f \n" % (dir1[0], dir1[1], dir1[2])
                    )
                    f.write("end\n")
                    f.write(
                        "cartesian %f %f %f \n" % (dir2[0], dir2[1], dir2[2])
                    )
                    f.write("end\n")
                    f.write("1\n")
                    # 4966 is hard coded for Ti
                    # NEED TO FIX THIS (probably by moving it to a lookup
                    # table inside OCEAN)
                    f.write(str(weight) + "\n")
                    f.close

        return {"pass": True, "errors": dict()}



from pathlib import Path
import xml.etree.ElementTree as ET
from warnings import warn

from monty.json import MSONable
from pymatgen.io.exciting import ExcitingInput

from lightshow.parameters._base import _BaseParameters
from lightshow.common.kpoints import GenericEstimatorKpoints
from lightshow.common.nbands import UnitCellVolumeEstimate
from lightshow import _get_SPECIES_DIRECTORY_from_environ

EXCITING_DEFAULT_GQMAX = 4.0

EXCITING_DEFAULT_CARDS = {
    "structure": {"speciespath": "./", "autormt": "true"},
    "groundstate": {
        "xctype": "GGA_PBE",
        "nempty": "200",
        "rgkmax": "9.0",
        "do": "fromscratch",
        "gmaxvr": "25",
        "lmaxmat": "10",
        "lmaxvr": "10",
        "lmaxapw": "10",
    },
    "xs": {
        "xstype": "BSE",
        "vkloff": "0.05 0.03 0.13",
        "nempty": "150",
        "gqmax": str(EXCITING_DEFAULT_GQMAX),
        "broad": "0.0327069",
        "tevout": "true",
        "tappinfo": "true",
        "energywindow": {"intv": "178.2 180.5", "points": "1000"},
        "screening": {"screentype": "full", "nempty": "150"},
        "BSE": {
            "xas": "true",
            "bsetype": "singlet",
            "nstlxas": "1 20",
            "distribute": "true",
            "eecs": "1000",
        },
        "qpointset": {"qpoint": {"text()": "0.0 0.0 0.0"}},
    },
}


class EXCITINGParameters(MSONable, _BaseParameters):
    """A one-stop-shop for all the different ways to modify input parameters
    for an Exciting calculation. This class is a lightweight wrapper for the
    ``ExcitingInput``, containing some extra information and methods used for
    writing the appropriate input files. Like most Pymatgen objects, the
    class is serializable via e.g. ``exciting_parameters.as_dict()``.

    Parameters
    ----------
    cards : dict
        A dictionary of of the cards to be control the paramters in the
        EXCITING calculations. It contains all the default setting in
        this workflow.

        .. code-block:: python

            cards = {
                "structure": {"speciespath": "./", "autormt": "true"},
                "groundstate": {
                    "xctype": "GGA_PBE",
                    "nempty": "200",
                    "rgkmax": "9.0",
                    "do": "fromscratch",
                    "gmaxvr": "25",
                    "lmaxmat": "10",
                    "lmaxvr": "10",
                    "lmaxapw": "10",
                },
                "xs": {
                    "xstype": "BSE",
                    "vkloff": "0.05 0.03 0.13",
                    "nempty": "150",
                    "gqmax": "4.0",
                    "broad": "0.0327069",
                    "tevout": "true",
                    "tappinfo": "true",
                    "energywindow": {"intv": "178.2 180.5", "points": "1000"},
                    "screening": {"screentype": "full", "nempty": "150"},
                    "BSE": {
                        "xas": "true",
                        "xasedge": "K",
                        "bsetype": "singlet",
                        "nstlxas": "1 20",
                        "distribute": "true",
                        "eecs": "1000",
                    },
                    "qpointset": {"qpoint": {"text()": "0.0 0.0 0.0"}},
                }
            }
    edge : str
        The XAS edge for the calculation
    species_directory : str
        A string contains the absolute path for species files. Usually, this
        folder comes with the exciting package, which can be obtained at
        http://exciting.wikidot.com/
    kpoints : lightshow.common.kpoints._BaseKpointsMethod
        The method for constructing the kpoints file from the structure. Should
        be a class with a ``__call__`` method defined. This method should take
        the structure as input and return a tuple corresponding to the kpoints
        density along each axis.
    nbands : lightshow.common.nbands._BaseNbandsMethod
        The method for determining the number of valence bands from the
        structure. Should be a class with a ``__call__`` method defined. This
        method should take the structure as input and return an integer: the
        number of valence bands to use in the calculation.
    gqmax : float
        :math:`|G+q|` cutoff of the plane wave expansion
    name : str
        The name of the calculation. Should likely always be ``"EXCITING"``.
    """

    @property
    def name(self):
        return self._name

    def __init__(
        self,
        cards=EXCITING_DEFAULT_CARDS,
        species_directory=None,
        plan=[
            "xsgeneigvec",
            "writepmatxs",
            "scrgeneigvec",
            "scrwritepmat",
            "screen",
            "scrcoulint",
            "exccoulint",
            "bse",
        ],
        edge="K",
        kpoints=GenericEstimatorKpoints(cutoff=16.0, max_radii=50.0),
        nbands=UnitCellVolumeEstimate(e_range=30.0),
        gqmax=EXCITING_DEFAULT_GQMAX,
        name="EXCITING",
    ):
        # Default cards
        self._cards = cards

        # species directory
        if species_directory is None:
            species_directory = _get_SPECIES_DIRECTORY_from_environ()
        if species_directory is None:
            warn(
                "species_directory not set, and SPECIES_DIRECTORY not in "
                "the current environment variables. The current/working "
                "folder will be used as default. Please make sure you copy "
                "all the corresponding species file into the working folder, "
                "e.g. where the input.xml file is generated."
            )
            species_directory = "./"
        self._species_directory = species_directory
        self._cards["structure"]["speciespath"] = self._species_directory
        # Update plan
        self._plan = plan

        # Set XAS edge
        self._edge = edge
        self._cards["xs"]["BSE"]["xasedge"] = self._edge
        # Method for determining the kmesh
        self._kpoints = kpoints
        # Method for determining number of bands
        self._nbands = nbands

        # Set gqmax
        self._gqmax = str(gqmax)
        self._cards["xs"]["gqmax"] = self._gqmax

        # Method for determining the kmesh
        self._name = name

    def write(self, target_directory, **kwargs):
        """Writes the input files for the provided structures and sites.

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the FEFF input files.
        **kwargs
            Must contain the ``structure_uc`` key (the
            :class:`pymatgen.core.structure.Structure` of interest) and the
            ``sites`` key (a list of int, where each int corresponds to the
            site index of the site to write).

        Returns
        -------
        dict
            A dictionary containing the status and errors key. In the case of
            EXCITING, there are no possible errors at this stage other than
            critical ones that would cause program termination, so the returned
            object is always ``{"pass": True, "errors": dict()}``.
        """

        structure = kwargs["structure_uc"]
        sites = kwargs["sites"]

        target_directory = Path(target_directory)
        target_directory.mkdir(exist_ok=True, parents=True)

        excitinginput = ExcitingInput(structure)
        # Estimate number of band
        nbands = self._nbands(structure)
        self._cards["xs"]["BSE"]["nstlxas"] = f"1 {nbands}"
        # Estimate number of kpoints
        kmesh = self._kpoints(structure)
        self._cards["groundstate"][
            "ngridk"
        ] = f"{kmesh[0]} {kmesh[1]} {kmesh[2]}"
        self._cards["xs"]["ngridk"] = f"{kmesh[0]} {kmesh[1]} {kmesh[2]}"
        self._cards["xs"]["ngridq"] = f"{kmesh[0]} {kmesh[1]} {kmesh[2]}"
        # Determine XAS species
        species = [structure[site].specie.symbol for site in sites]

        for i, specie in enumerate(
            sorted(structure.types_of_species, key=lambda el: el.X)
        ):
            if specie.symbol == species[0]:
                self._cards["xs"]["BSE"]["xasspecies"] = str(i + 1)

        for site, specie in zip(sites, species):
            path = target_directory / f"{site:03}_{specie}"
            path.mkdir(exist_ok=True, parents=True)

            filepath_xas = path / "input.xml"
            self._cards["xs"]["BSE"]["xasatom"] = str(site + 1)
            # add the plans using the tree
            root = excitinginput.write_etree(
                "primitive", bandstr=False, **self._cards
            )
            tree = ET.ElementTree(root)
            xs_loc = root.find("xs")
            plan = ET.Element("plan")
            xs_loc.insert(-1, plan)
            ploc = xs_loc.find("plan")
            for task in self._plan:
                ET.SubElement(ploc, "doonly", task=task)

            excitinginput._indent(root)
            tree.write(filepath_xas)

        return {"pass": True, "errors": dict()}



"""
Basic tests for exciting.
"""
import re
from pathlib import Path

from pymatgen.core.structure import Structure
from xml.etree import ElementTree

from lightshow.parameters.exciting import EXCITINGParameters


def test_default_parameters(mp_Structure_mp390: Structure, tmp_path: Path):
    """Test the default parameters of exciting."""
    exciting_parameters = EXCITINGParameters()
    assert exciting_parameters.name == "EXCITING"
    write_results = exciting_parameters.write(
        tmp_path, structure_uc=mp_Structure_mp390, sites=[1]
    )
    assert write_results == {"pass": True, "errors": {}}
    ti_path = tmp_path / "001_Ti"
    assert ti_path.exists(), "Directory for absorbing site was not generated."
    input_file = ti_path / "input.xml"
    assert input_file.exists(), "Input file not found."

    root = ElementTree.parse(input_file).getroot()

    assert root.tag == "input"

    subelements = list(root)
    assert len(subelements) == 4

    title_xml = subelements[0]
    assert title_xml.tag == "title"
    assert title_xml.keys() == []
    assert title_xml.text == "Ti2 O4"

    structure_xml = subelements[1]
    assert structure_xml.tag == "structure"
    assert set(structure_xml.keys()) == {"speciespath", "autormt"}
    assert structure_xml.get("speciespath") == "./", "species path set to ./"
    assert structure_xml.get("autormt") == "true"

    assert (
        len(list(structure_xml)) == 3
    ), "Expect structure tree to have 3 sub-elements"
    crystal_xml = structure_xml[0]
    assert crystal_xml.tag == "crystal", "First subtree is crystal"
    assert crystal_xml.items() == [("scale", "1.8897543760313331")]

    lattice_vectors = list(crystal_xml)
    assert len(lattice_vectors) == 3, "Always expect three lattice vectors"
    assert lattice_vectors[0].items() == [], "Lattice vectors have no items"
    assert (
        lattice_vectors[0].text
        == "     -1.90135500       1.90135500       4.87387600"
    ), "Lattice vector `a` differs from input"
    assert (
        lattice_vectors[1].text
        == "      1.90135500      -1.90135500       4.87387600"
    ), "Lattice vector `b` differs from input"
    assert (
        lattice_vectors[2].text
        == "      1.90135500       1.90135500      -4.87387600"
    ), "Lattice vector `c` differs from input"

    species_ti_xml = structure_xml[1]
    assert species_ti_xml.tag == "species", "Second subtree is species"

    species_o_xml = structure_xml[2]
    assert species_o_xml.tag == "species", "Third subtree is species"

    assert species_ti_xml.items() == [
        ("speciesfile", "Ti.xml")
    ], "species is inconsistent"
    assert species_o_xml.items() == [
        ("speciesfile", "O.xml")
    ], "species is inconsistent"

    atoms_ti = list(species_ti_xml)
    assert len(atoms_ti) == 2, "Number of Ti atoms is wrong"
    assert atoms_ti[0].items() == [
        ("coord", "      0.75000000       0.25000000       0.50000000")
    ], "Coordinate of first Ti differs to input"
    assert atoms_ti[1].items() == [
        ("coord", "      0.00000000       0.00000000       0.00000000")
    ], "Coordinate of second Ti differs to input"

    atoms_o = list(species_o_xml)
    assert len(atoms_o) == 4, "Number of O atoms is wrong"
    assert atoms_o[0].items() == [
        ("coord", "      0.95616300       0.45616300       0.50000000")
    ], "Coordinate of first O differs to input"
    assert re.match(
        r" {6}0\.79383700 {7}0\.79383700 {7}[01]\.00000000",
        atoms_o[1].items()[0][1],
    ), "Coordinate of second O differs to input"
    assert atoms_o[2].items() == [
        ("coord", "      0.54383700       0.04383700       0.50000000")
    ], "Coordinate of third O differs to input"
    assert atoms_o[3].items() == [
        ("coord", "      0.20616300       0.20616300       0.00000000")
    ], "Coordinate of fourth O differs to input"

    groundstate_xml = subelements[2]
    assert groundstate_xml.tag == "groundstate"
    assert (
        len(groundstate_xml.keys()) == 9
    ), "Excpect groundstate to have 9 attributes."
    assert set(groundstate_xml.keys()) == {
        "rgkmax",
        "do",
        "ngridk",
        "xctype",
        "nempty",
        "gmaxvr",
        "lmaxmat",
        "lmaxvr",
        "lmaxapw",
    }, "Excpected groundstate keys differ from found keys."
    assert groundstate_xml.get("rgkmax") == "9.0"
    assert groundstate_xml.get("do") == "fromscratch"
    assert groundstate_xml.get("ngridk") == "5 5 6"
    assert groundstate_xml.get("xctype") == "GGA_PBE"
    assert groundstate_xml.get("nempty") == "200"
    assert groundstate_xml.get("gmaxvr") == "25"
    assert groundstate_xml.get("lmaxmat") == "10"
    assert groundstate_xml.get("lmaxvr") == "10"
    assert groundstate_xml.get("lmaxapw") == "10"

    xs_xml = subelements[3]
    assert xs_xml.tag == "xs"
    assert len(xs_xml.keys()) == 9, "Excpect xs to have 9 attributes."
    assert set(xs_xml.keys()) == {
        "broad",
        "ngridk",
        "xstype",
        "nempty",
        "gqmax",
        "vkloff",
        "tevout",
        "tappinfo",
        "ngridq",
    }, "Excpected xs keys differ from found keys."
    assert xs_xml.get("broad") == "0.0327069"
    assert xs_xml.get("ngridk") == "5 5 6"
    assert xs_xml.get("xstype") == "BSE"
    assert xs_xml.get("nempty") == "150"
    assert xs_xml.get("gqmax") == "4.0"
    assert xs_xml.get("vkloff") == "0.05 0.03 0.13"
    assert xs_xml.get("tevout") == "true"
    assert xs_xml.get("tappinfo") == "true"
    assert xs_xml.get("ngridq") == "5 5 6"

    xs_subelements = list(xs_xml)
    assert len(xs_subelements) == 5, "Expect xs tree to have 5 sub-elements"

    energywindow_xml = xs_subelements[0]
    assert (
        energywindow_xml.tag == "energywindow"
    ), "First xs subelement should be energywindow"
    assert set(energywindow_xml.keys()) == {"intv", "points"}
    assert energywindow_xml.get("intv") == "178.2 180.5"
    assert energywindow_xml.get("points") == "1000"

    screening_xml = xs_subelements[1]
    assert (
        screening_xml.tag == "screening"
    ), "Second xs subelement should be screening"
    assert set(screening_xml.keys()) == {"screentype", "nempty"}
    assert screening_xml.get("screentype") == "full"
    assert screening_xml.get("nempty") == "150"

    bse_xml = xs_subelements[2]
    assert bse_xml.tag == "BSE", "Third xs subelement should be BSE"
    assert set(bse_xml.keys()) == {
        "bsetype",
        "xas",
        "xasedge",
        "nstlxas",
        "distribute",
        "eecs",
        "xasspecies",
        "xasatom",
    }
    assert bse_xml.get("bsetype") == "singlet"
    assert bse_xml.get("xas") == "true"
    assert bse_xml.get("xasedge") == "K"
    assert bse_xml.get("nstlxas") == "1 59"
    assert bse_xml.get("distribute") == "true"
    assert bse_xml.get("eecs") == "1000"
    assert bse_xml.get("xasspecies") == "1"
    assert bse_xml.get("xasatom") == "2"

    plan_xml = xs_subelements[3]
    assert plan_xml.tag == "plan", "Fourth xs subelement should be plan"
    assert plan_xml.items() == []
    doonlys = list(plan_xml)
    assert len(doonlys) == 8, "Excpect 8 plan elements"
    assert doonlys[0].tag == "doonly"
    assert doonlys[0].items() == [("task", "xsgeneigvec")]
    assert doonlys[1].tag == "doonly"
    assert doonlys[1].items() == [("task", "writepmatxs")]
    assert doonlys[2].tag == "doonly"
    assert doonlys[2].items() == [("task", "scrgeneigvec")]
    assert doonlys[3].tag == "doonly"
    assert doonlys[3].items() == [("task", "scrwritepmat")]
    assert doonlys[4].tag == "doonly"
    assert doonlys[4].items() == [("task", "screen")]
    assert doonlys[5].tag == "doonly"
    assert doonlys[5].items() == [("task", "scrcoulint")]
    assert doonlys[6].tag == "doonly"
    assert doonlys[6].items() == [("task", "exccoulint")]
    assert doonlys[7].tag == "doonly"
    assert doonlys[7].items() == [("task", "bse")]

    qpointset_xml = xs_subelements[4]
    assert (
        qpointset_xml.tag == "qpointset"
    ), "Fifth xs subelement should be qpointset"
    assert qpointset_xml.items() == []
    qpoints = list(qpointset_xml)
    assert len(qpoints) == 1, "Excpect 1 qpoint"
    assert qpoints[0].tag == "qpoint"
    assert qpoints[0].items() == []
    assert qpoints[0].text == "0.0 0.0 0.0"


def test_exciting_own_parameters(tmp_path: Path, mp_Structure_mp390: Structure):
    own_cards = {
        "structure": {},
        "groundstate": {
            "nempty": "42",
        },
        "xs": {
            "xstype": "BSE",
            "nempty": "43",
            "energywindow": {"points": "44"},
            "BSE": {},
        },
    }

    exciting_parameters = EXCITINGParameters(
        own_cards, "keks", ["piep", "blub"], name="Test"
    )
    assert exciting_parameters.name == "Test"

    write_results = exciting_parameters.write(
        tmp_path, structure_uc=mp_Structure_mp390, sites=[1, 2, 3]
    )
    assert write_results == {"pass": True, "errors": {}}

    o1_path = tmp_path / "002_O"
    assert o1_path.exists(), "Directory for absorbing site 2 was not generated."
    input_file_o1 = o1_path / "input.xml"
    assert input_file_o1.exists(), "Input file not found."

    o2_path = tmp_path / "003_O"
    assert o2_path.exists(), "Directory for absorbing site 3 was not generated."
    input_file_o2 = o2_path / "input.xml"
    assert input_file_o2.exists(), "Input file not found."

    ti_path = tmp_path / "001_Ti"
    assert ti_path.exists(), "Directory for absorbing site 1 was not generated."
    input_file_ti = ti_path / "input.xml"
    assert input_file_ti.exists(), "Input file not found."

    root = ElementTree.parse(input_file_ti).getroot()

    assert root.tag == "input"

    subelements = list(root)
    assert len(subelements) == 4

    structure_xml = subelements[1]
    assert structure_xml.tag == "structure"
    assert (
        structure_xml.get("speciespath") == "keks"
    ), "species path set to keks"

    groundstate_xml = subelements[2]
    assert groundstate_xml.tag == "groundstate"
    assert (
        len(groundstate_xml.keys()) == 2
    ), "Excpect groundstate to have 2 attributes."
    assert set(groundstate_xml.keys()) == {
        "ngridk",
        "nempty",
    }, "Excpected groundstate keys differ from found keys."
    assert groundstate_xml.get("ngridk") == "5 5 6"
    assert groundstate_xml.get("nempty") == "42"

    xs_xml = subelements[3]
    assert xs_xml.tag == "xs"
    assert len(xs_xml.keys()) == 5, "Excpect xs to have 4 attributes."
    assert set(xs_xml.keys()) == {
        "ngridk",
        "nempty",
        "ngridq",
        "xstype",
        "gqmax",
    }, "Excpected xs keys differ from found keys."
    assert xs_xml.get("gqmax") == "4.0"
    assert xs_xml.get("ngridk") == "5 5 6"
    assert xs_xml.get("xstype") == "BSE"
    assert xs_xml.get("nempty") == "43"
    assert xs_xml.get("ngridq") == "5 5 6"

    xs_subelements = list(xs_xml)
    assert len(xs_subelements) == 3, "Expect xs tree to have 3 sub-elements"

    energywindow_xml = xs_subelements[0]
    assert (
        energywindow_xml.tag == "energywindow"
    ), "First xs subelement should be energywindow"
    assert set(energywindow_xml.keys()) == {"points"}
    assert energywindow_xml.get("points") == "44"

    plan_xml = xs_subelements[1]
    assert plan_xml.tag == "plan", "Second xs subelement should be plan"
    assert plan_xml.items() == []
    doonlys = list(plan_xml)
    assert len(doonlys) == 2, "Excpect 2 plan elements"
    assert doonlys[0].tag == "doonly"
    assert doonlys[0].items() == [("task", "piep")]
    assert doonlys[1].tag == "doonly"
    assert doonlys[1].items() == [("task", "blub")]

    bse_xml = xs_subelements[2]
    assert bse_xml.tag == "BSE", "Third xs subelement should be BSE"
    assert set(bse_xml.keys()) == {
        "nstlxas",
        "xasedge",
        "xasspecies",
        "xasatom",
    }
    assert bse_xml.get("nstlxas") == "1 59"
    assert bse_xml.get("xasspecies") == "1"
    assert bse_xml.get("xasatom") == "2"

    bse_xml_o1 = list(list(ElementTree.parse(input_file_o1).getroot())[3])[2]
    assert bse_xml_o1.get("xasspecies") == "1"
    assert bse_xml_o1.get("xasatom") == "3"

    bse_xml_o2 = list(list(ElementTree.parse(input_file_o2).getroot())[3])[2]
    assert bse_xml_o2.get("xasspecies") == "1"
    assert bse_xml_o2.get("xasatom") == "4"



"""Helper module for making the API a bit cleaner when dealing with default
parameters."""

# Note that pre-commit + .flake8 ignores this file since it's just a bunch of
# imports...

from lightshow.parameters.feff import FEFF_DEFAULT_CARDS

from lightshow.parameters.vasp import (
    VASP_INCAR_DEFAULT_NEUTRAL_POTENTIAL,
    VASP_INCAR_DEFAULT_COREHOLE_POTENTIAL,
    VASP_POTCAR_DEFAULT_ELEMENT_MAPPING,
)

from lightshow.parameters.ocean import OCEAN_DEFAULT_CARDS

from lightshow.parameters.xspectra import XSPECTRA_DEFAULT_CARDS

from lightshow.parameters.exciting import EXCITING_DEFAULT_CARDS



from pathlib import Path
import pytest
import sys

from pymatgen.ext.matproj import MPRestError
from warnings import warn

from lightshow import (
    Database,
    FEFFParameters,
    VASPParameters,
    OCEANParameters,
    XSpectraParameters,
    EXCITINGParameters,
)
from lightshow.defaults import VASP_INCAR_DEFAULT_COREHOLE_POTENTIAL

# Helper testing files
sys.path.append(str(Path(__file__).parent.resolve() / Path("helpers")))
from geometry import consistency_check  # noqa


def test_database_from_disk(database_from_file, test_structure_names):
    dat = database_from_file
    dat.initialize_supercells(9.0)
    dat.initialize_inequivalent_sites()
    assert set(dat.structures.keys()) == set(test_structure_names)
    assert set(dat.metadata.keys()) == set(test_structure_names)
    assert set(dat.supercells.keys()) == set(test_structure_names)


def test_from_materials_project(test_structure_names):
    try:
        dat = Database.from_materials_project(test_structure_names)
        dat.initialize_supercells(9.0)
        assert set(dat.structures.keys()) == set(test_structure_names)
        assert set(dat.supercells.keys()) == set(test_structure_names)
    except MPRestError:
        warn("MPRestError during pulling data")


@pytest.mark.parametrize(
    "mpid",
    [
        "mp-390",
        "mvc-11115",
        "mp-1215",
        "mp-1840",
        "mp-2657",
        "mp-2664",
        "mp-430",
        "mp-458",
        "mp-10734",
    ],
)
def test_write(
    mpid,
    dummy_potcar_file_directory,
    dummy_psp_file_directory,
    dummy_chpsp_file_directory,
    database_from_file,
    tmp_path,
):

    # Load it all in
    dat = database_from_file
    dat.initialize_supercells(9.0)
    dat.initialize_inequivalent_sites()
    dat._supercells = {mpid: dat.supercells[mpid]}
    dat._structures = {mpid: dat.structures[mpid]}
    dat._metadata = {mpid: dat.metadata[mpid]}

    target = Path(tmp_path) / Path("iama") / Path("destination")
    target.mkdir(exist_ok=True, parents=True)

    # Write it (note these are Ti-O compounds)
    feff_parameters = FEFFParameters(
        cards={
            "S02": "0",
            "COREHOLE": "RPA",
            "CONTROL": "1 1 1 1 1 1",
            "XANES": "4 0.04 0.1",
            "SCF": "7.0 0 100 0.2 3",
            "FMS": "9.0 0",
            "EXCHANGE": "0 0.0 0.0 2",
            "RPATH": "-1",
        },
        edge="K",
        radius=6.0,
        spectrum="XANES",
        name="FEFF",
    )
    vasp_params_corehole = VASPParameters(
        incar=VASP_INCAR_DEFAULT_COREHOLE_POTENTIAL,
        potcar_directory=dummy_potcar_file_directory,
        force_spin_unpolarized=False,
    )
    ocean_params = OCEANParameters(edge="K")
    exciting_params = EXCITINGParameters(edge="K")
    xspectra_params = XSpectraParameters(
        psp_directory=dummy_psp_file_directory,
        psp_cutoff_table="mock_cutoff_table.json",
        chpsp_directory=dummy_chpsp_file_directory,
        edge="K",
    )

    dat.write(
        target,
        absorbing_atoms=["Ti", "O"],
        options=[
            feff_parameters,
            vasp_params_corehole,
            ocean_params,
            exciting_params,
            xspectra_params,
        ],
    )

    # Assert geometires
    consistency_check(target / Path(mpid), rounding=3)



from pathlib import Path
from functools import lru_cache

import numpy as np
from pymatgen.core.structure import IStructure
from pymatgen.io.exciting import ExcitingInput


ATOMIC_NUMBERS = {
    "H": 1,
    "He": 2,
    "Li": 3,
    "Be": 4,
    "B": 5,
    "C": 6,
    "N": 7,
    "O": 8,
    "F": 9,
    "Ne": 10,
    "Na": 11,
    "Mg": 12,
    "Al": 13,
    "Si": 14,
    "P": 15,
    "S": 16,
    "Cl": 17,
    "Ar": 18,
    "K": 19,
    "Ca": 20,
    "Sc": 21,
    "Ti": 22,
    "V": 23,
    "Cr": 24,
    "Mn": 25,
    "Fe": 26,
    "Co": 27,
    "Ni": 28,
    "Cu": 29,
    "Zn": 30,
    "Ga": 31,
    "Ge": 32,
    "As": 33,
    "Se": 34,
    "Br": 35,
    "Kr": 36,
    "Rb": 37,
    "Sr": 38,
    "Y": 39,
    "Zr": 40,
    "Nb": 41,
    "Mo": 42,
    "Tc": 43,
    "Ru": 44,
    "Rh": 45,
    "Pd": 46,
    "Ag": 47,
    "Cd": 48,
    "In": 49,
    "Sn": 50,
    "Sb": 51,
    "Te": 52,
    "I": 53,
    "Xe": 54,
    "Cs": 55,
    "Ba": 56,
    "La": 57,
    "Ce": 58,
    "Pr": 59,
    "Nd": 60,
    "Pm": 61,
    "Sm": 62,
    "Eu": 63,
    "Gd": 64,
    "Tb": 65,
    "Dy": 66,
    "Ho": 67,
    "Er": 68,
    "Tm": 69,
    "Yb": 70,
    "Lu": 71,
    "Hf": 72,
    "Ta": 73,
    "W": 74,
    "Re": 75,
    "Os": 76,
    "Ir": 77,
    "Pt": 78,
    "Au": 79,
    "Hg": 80,
    "Tl": 81,
    "Pb": 82,
    "Bi": 83,
    "Po": 84,
    "At": 85,
    "Rn": 86,
    "Fr": 87,
    "Ra": 88,
    "Ac": 89,
    "Th": 90,
    "Pa": 91,
    "U": 92,
    "Np": 93,
    "Pu": 94,
    "Am": 95,
    "Cm": 96,
    "Bk": 97,
    "Cf": 98,
    "Es": 99,
    "Fm": 100,
    "Md": 101,
    "No": 102,
    "Lr": 103,
    "Rf": 104,
    "Db": 105,
    "Sg": 106,
    "Bh": 107,
    "Hs": 108,
    "Mt": 109,
    "Ds": 110,
    "Rg": 111,
    "Cn": 112,
    "Nh": 113,
    "Fl": 114,
    "Mc": 115,
    "Lv": 116,
    "Ts": 117,
    "Og": 118,
    "Uue": 119,
}
ATOMIC_NUMBERS = {value: key for key, value in ATOMIC_NUMBERS.items()}


def read_FEFF_geometry(path, rounding=4):

    path = Path(path) / "feff.inp"

    with open(path, "r") as f:
        feff_lines = f.readlines()
    where_atoms = [
        ii for ii, xx in enumerate(feff_lines) if xx.startswith("ATOMS")
    ]
    assert len(where_atoms) == 1
    feff_lines = feff_lines[where_atoms[0] + 3 : -1]
    feff_lines = [xx.split() for xx in feff_lines]

    atoms = [xx[4] for xx in feff_lines]
    distances = np.round([float(xx[5]) for xx in feff_lines], rounding)

    return {"atoms": atoms[1:], "distances": distances[1:]}


def read_VASP_geometry(path, neighbor_radius=10.0, rounding=4):

    path = Path(path) / "POSCAR"

    # VASP POSCAR files are easy, only need data after line 8
    structure = IStructure.from_file(path)
    # vasp_coordinates = np.array([site.frac_coords for site in vasp_structure])
    # vasp_atoms = [site.specie.symbol for site in vasp_structure]

    neigh = structure.get_neighbors(structure[0], r=neighbor_radius)
    tmp = [[xx.nn_distance, str(xx.specie)] for xx in neigh]
    tmp.sort(key=lambda xx: xx[0])

    return {
        "atoms": [xx[1] for xx in tmp],
        "distances": np.round([xx[0] for xx in tmp], rounding),
    }


def read_OCEAN_geometry(path, neighbor_radius=10.0, rounding=4):

    path = Path(path) / "ocean.in"

    absorber = str(path.parts[-2])

    with open(path, "r") as f:
        ocean_lines = f.readlines()
    ocean_lines = [xx.strip() for xx in ocean_lines]

    # Parse the ocean input file, it's kinda annoying...
    ocean_lines = (
        " ".join(ocean_lines).replace("{", "").replace("}", "|")[:-1].strip()
    )
    ocean_lines = ocean_lines.split("|")
    ocean_lines = [xx.strip() for xx in ocean_lines]

    geometry = None
    mat = None
    znucl = None
    typat = None

    for line in ocean_lines:
        split = line.split()
        key = split[0]

        if key == "xred":
            geometry = np.array(split[1:], dtype=float).reshape(-1, 3)
        elif key == "rprim":
            mat = np.array(split[1:], dtype=float).reshape(3, 3)
        elif key == "znucl":
            znucl = np.array(split[1:], dtype=int)
        elif key == "typat":
            typat = np.array(split[1:], dtype=int)

    assert geometry is not None
    assert mat is not None
    assert znucl is not None
    assert typat is not None

    atoms = [ATOMIC_NUMBERS[znucl[xx - 1]] for xx in typat]

    structure = IStructure(lattice=mat, species=atoms, coords=geometry)

    # Need to deal with this. OCEAN's indices of relevance are not necessarily 0
    absorbing_sites = [ii for ii, atom in enumerate(atoms) if atom == absorber]
    return_list = []
    for site in absorbing_sites:
        neigh = structure.get_neighbors(structure[site], r=neighbor_radius)
        tmp = [[xx.nn_distance, str(xx.specie)] for xx in neigh]
        tmp.sort(key=lambda xx: xx[0])
        return_list.append(
            {
                "atoms": [xx[1] for xx in tmp],
                "distances": np.round([xx[0] for xx in tmp], rounding),
            }
        )
    return return_list


def read_XSpectra_geometry(path, neighbor_radius=10.0, rounding=4):

    path = Path(path) / "es.in"

    with open(path, "r") as f:
        xspectra_lines = f.readlines()
    xspectra_lines = [xx.strip() for xx in xspectra_lines]

    # Parse the Xspectra lines... also a pain, just like OCEAN
    atoms = []
    geom = []
    mat = None
    absorber = None
    ii = 0
    while ii < len(xspectra_lines):
        line = xspectra_lines[ii]
        if "ATOMIC_POSITIONS" in line:
            ii += 1
            jj = 0
            while True:
                line = xspectra_lines[ii]
                line_split = line.split()
                atom = line_split[0]
                position = line_split[1:]
                if "+" in atom:
                    absorber = jj
                    atom = atom[:-1]  # Remove the "+"
                if atom not in ATOMIC_NUMBERS.values():
                    break
                atoms.append(atom)
                geom.append(position)

                ii += 1
                jj += 1

        if "CELL_PARAMETERS" in line:
            mat = np.array(
                [xx.split() for xx in xspectra_lines[ii + 1 : ii + 4]],
                dtype=float,
            )
            break

        ii += 1

    assert absorber is not None

    geom = np.array(geom, dtype=float)

    structure = IStructure(lattice=mat, species=atoms, coords=geom)

    neigh = structure.get_neighbors(structure[absorber], r=neighbor_radius)
    tmp = [[xx.nn_distance, str(xx.specie)] for xx in neigh]
    tmp.sort(key=lambda xx: xx[0])

    return {
        "atoms": [xx[1] for xx in tmp],
        "distances": np.round([xx[0] for xx in tmp], rounding),
    }


def read_EXCITING_geometry(path, neighbor_radius=10.0, rounding=4):

    path = Path(path) / "input.xml"
    structure = ExcitingInput.from_file(path).structure

    # Get the absorbing index
    with open(path, "r") as f:
        exciting_lines = f.readlines()
    exciting_lines = [xx.strip() for xx in exciting_lines]
    absorbing_line = [xx for xx in exciting_lines if "xasatom" in xx]
    assert len(absorbing_line) == 1
    absorbing_line = absorbing_line[0].split()
    absorbing_line = [xx for xx in absorbing_line if "xasatom" in xx]
    assert len(absorbing_line) == 1
    absorbing_line = absorbing_line[0]
    absorber = int(absorbing_line.split("=")[1].replace('"', "")) - 1

    neigh = structure.get_neighbors(structure[absorber], r=neighbor_radius)
    tmp = [[xx.nn_distance, str(xx.specie)] for xx in neigh]
    tmp.sort(key=lambda xx: xx[0])

    return {
        "atoms": [xx[1] for xx in tmp],
        "distances": np.round([xx[0] for xx in tmp], rounding),
    }


@lru_cache(maxsize=16)
def _read_OCEAN_geometry(path, neighbor_radius, rounding):
    return read_OCEAN_geometry(
        path, neighbor_radius=neighbor_radius, rounding=rounding
    )


def consistency_check(
    path, rounding=3, first_n_distances=10, neighbor_radius=10.0
):
    """Summary

    Parameters
    ----------
    path : os.PathLike
        A path to a particular materials directory.
    rounding : int, optional
        The number of decimal points to round the distances to.
    first_n_distances : int, optional
        The first n distances are taken to do the consistency check. These
        distances are sorted in the order of closest to furthest to the
        absorbing atom.
    """

    atom_dirs_FEFF = sorted(list((Path(path) / "FEFF").iterdir()))
    atom_dirs_VASP = sorted(list((Path(path) / "VASP").iterdir()))
    atom_dirs_XSpectra = sorted(list((Path(path) / "XSpectra").iterdir()))
    atom_dirs_EXCITING = sorted(list((Path(path) / "EXCITING").iterdir()))

    l1 = [xx.name for xx in atom_dirs_FEFF]
    l2 = [xx.name for xx in atom_dirs_VASP]
    l3 = [xx.name for xx in atom_dirs_XSpectra]
    l4 = [xx.name for xx in atom_dirs_EXCITING]
    if not l1 == l2 == l4:
        raise AssertionError(f"\n{l1}\n{l2}\n{l4}")

    # Note that XSpectra comes with quite a few extra files...
    remaining_length = len(l3) - len(l2)
    if remaining_length != len(set(l3) - set(l2)):
        raise AssertionError(f"\n{l2}\n{l3}")

    for path_FEFF, path_VASP, path_XSpectra, path_EXCITING in zip(
        atom_dirs_FEFF, atom_dirs_VASP, atom_dirs_XSpectra, atom_dirs_EXCITING
    ):
        data_FEFF = read_FEFF_geometry(path_FEFF, rounding=rounding)
        data_VASP = read_VASP_geometry(
            path_VASP, neighbor_radius=neighbor_radius, rounding=rounding
        )
        data_XSpectra = read_XSpectra_geometry(
            path_XSpectra, neighbor_radius=neighbor_radius, rounding=rounding
        )
        data_EXCITING = read_EXCITING_geometry(
            path_EXCITING, neighbor_radius=neighbor_radius, rounding=rounding
        )

        a1 = data_FEFF["atoms"][:first_n_distances]
        a2 = data_VASP["atoms"][:first_n_distances]
        a3 = data_XSpectra["atoms"][:first_n_distances]
        a4 = data_EXCITING["atoms"][:first_n_distances]
        assert a1 == a2 == a3 == a4, f"\n{a1}\n{a2}\n{a3}\n{a4}"

        d1 = data_FEFF["distances"][:first_n_distances]
        d2 = data_VASP["distances"][:first_n_distances]
        d3 = data_XSpectra["distances"][:first_n_distances]
        d4 = data_EXCITING["distances"][:first_n_distances]
        if not np.allclose(d1, d2):
            raise AssertionError(f"\n{d1}\n{d2}")
        if not np.allclose(d2, d3):
            raise AssertionError(f"\n{d2}\n{d3}")
        if not np.allclose(d3, d4):
            raise AssertionError(f"\n{d3}\n{d4}")

        absorber = str(path_FEFF.name).split("_")[1]
        path_OCEAN = Path(path) / "OCEAN" / absorber

        all_data_OCEAN = _read_OCEAN_geometry(
            path_OCEAN, neighbor_radius=neighbor_radius, rounding=rounding
        )

        ocean_checks = []
        for data_OCEAN in all_data_OCEAN:

            if not (
                data_OCEAN["atoms"][:first_n_distances]
                == data_VASP["atoms"][:first_n_distances]
            ):
                ocean_checks.append(False)
            else:
                ocean_checks.append(True)

            if not np.allclose(
                data_OCEAN["distances"][:first_n_distances],
                data_VASP["distances"][:first_n_distances],
            ):
                ocean_checks.append(False)
            else:
                ocean_checks.append(True)

        assert any(ocean_checks)



from os import environ


# DO NOT CHANGE BELOW ---------------------------------------------------------
# This is replaced at build time automatically during deployment and
# installation. Replacing anything will mess that up and crash the entire
# build.
__version__ = '0.0.14'
# DO NOT CHANGE ABOVE ---------------------------------------------------------


def _get_API_key_from_environ():
    """Checks for an environment variable PMG_API_KEY. If does not exist,
    returns None.

    Returns
    -------
    str
    """

    return environ.get("PMG_API_KEY", None)


def _get_POTCAR_DIRECTORY_from_environ():
    """Checks for an environment variable VASP_POTCAR_DIRECTORY. If does not
    exist, returns None.

    Returns
    -------
    str
    """

    return environ.get("VASP_POTCAR_DIRECTORY", None)


def _get_CHPSP_DIRECTORY_from_environ():
    """Checks for an environment variable XS_CHPSP_DIRECTORY. If does not
    exist, returns None.

    Returns
    -------
    str
    """

    return environ.get("XS_CHPSP_DIRECTORY", None)


def _get_PSP_DIRECTORY_from_environ():
    """Checks for an environment variable XS_PSP_DIRECTORY. If does not
    exist, returns None.

    Returns
    -------
    str
    """

    return environ.get("XS_PSP_DIRECTORY", None)


def _get_SPECIES_DIRECTORY_from_environ():
    """Checks for an environment varialble SPECIES_DIRECRORY. If does not
    exist, returns None.

    Returns
    -------
    str
    """

    return environ.get("SPECIES_DIRECTORY", None)


from lightshow.database import Database  # noqa
from lightshow.parameters.feff import FEFFParameters  # noqa
from lightshow.parameters.vasp import VASPParameters  # noqa
from lightshow.parameters.exciting import EXCITINGParameters  # noqa
from lightshow.parameters.ocean import OCEANParameters  # noqa
from lightshow.parameters.xspectra import XSpectraParameters  # noqa



"""Contains all relevant classes and functions for constructing databases of
various materials and clusters. This includes pulling and processing data from
the `Materials Project Database <https://materialsproject.org/>`_, as well as
utilizing existing data the user may have on their hard drive."""

from datetime import datetime
import json
import os
from pathlib import Path
from shutil import copy2
from warnings import warn

from monty.json import MSONable
from pymatgen.core.structure import Structure
from pymatgen.ext.matproj import MPRester, MPRestError
from tqdm import tqdm

from lightshow import _get_API_key_from_environ
from lightshow import pymatgen_utils


def _delete_common_strings(old_list_of_strings):

    list_of_strings = [str(Path(xx).parent) for xx in old_list_of_strings]
    list_of_names = [str(Path(xx).name) for xx in old_list_of_strings]

    commonprefix = os.path.commonprefix(list_of_strings)
    new_p = [x[len(commonprefix) :] for x in list_of_strings]

    # Reverse every string in the list then do it again
    list_of_strings_reversed = [xx[::-1] for xx in new_p]
    commonsuffix = os.path.commonprefix(list_of_strings_reversed)
    new_p = [x[len(commonsuffix) :] for x in list_of_strings_reversed]

    final_p = [xx[::-1] for xx in new_p]

    return [
        str(Path(xx) / Path(name)) for xx, name in zip(final_p, list_of_names)
    ]


def _fetch_from_MP(mpr, mpid, metadata_keys):
    """Uses the provided MPID to fetch the structure data.

    Parameters
    ----------
    mpr : pymatgen.ext.matproj.MPRester
        Interface to the Materials Project REST API.
    mpid : str
        The specific mpid to pull.
    metadata_keys : list of str, optional
        The Materials Project metadata contains a huge amount of information.
        If not ``None``, these are the only keys that are kept in the pulled
        metadata object.

    Returns
    -------
    tuple
        The structure (:class:`pymatgen.core.structure.Structure`) of interest
        and the specified metadata.
    """

    metadata = mpr.get_doc(mpid)

    if metadata_keys is not None:
        metadata = {key: metadata[key] for key in metadata_keys}
    metadata["downloaded_at"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    # The structure is precisely in the data pulled from get_doc:
    structure = Structure.from_dict(metadata.pop("structure"))

    return structure, metadata


def _from_mpids_list(mpids, api_key, metadata_keys, verbose=True):
    """Makes one large API call to the Materials Project database and pulls the
    relevant structural files given a list of Materials Project ID's (mpids).

    Parameters
    ----------
    mpids : list of str
        List of Materials Project IDs to pull.
    api_key : str, optional
        Materials Project API key.
    metadata_keys : bool, optional
        If True, will only
    verbose : bool, optional
        If True, will use tqdm to print a progress bar.

    Returns
    -------
    dict
        A dictionary containing the structures (with keys as the mpids) and
        metadata (with the same keys).
    """

    structures = dict()
    metadatas = dict()
    with MPRester(api_key) as mpr:
        for mpid in tqdm(mpids, disable=not verbose):
            try:
                structure, metadata = _fetch_from_MP(mpr, mpid, metadata_keys)
            except MPRestError as error:
                warn(f"MPRestError pulling mpid={mpid}, error: {error}")
                continue
            structures[mpid] = structure.get_primitive_structure()
            metadatas[mpid] = metadata

    return {"structures": structures, "metadata": metadatas}


def _get_api_key(api_key):
    if api_key is None:
        api_key = _get_API_key_from_environ()
    if api_key is None:
        raise ValueError(f"Invalid API key {api_key}")
    return api_key


class Database(MSONable):
    """Contains all materials and metadata for some database."""

    def cleanup_paths(self):
        """When loading data from disk, paths can become very repetitive. This
        method strips all common prefixes and suffixes from the keys of the
        structures and metadata properties."""

        old_keys = list(self._structures.keys())
        new_keys = _delete_common_strings(old_keys)
        for old_key, new_key in zip(old_keys, new_keys):
            self._structures[new_key] = self._structures.pop(old_key)
            self._metadata[new_key] = self._metadata.pop(old_key)

            if self._supercells_initialized:
                self._supercells[new_key] = self._supercells.pop(old_key)

    @classmethod
    def from_files(cls, root, filename="CONTCAR", cleanup_paths=True):
        """Searches for files matching the provided ``filename``, which can
        include wildcards, and assumes those files are structural files in CIF
        format. The names/ids of these files is given by the full directory
        structure where that file was found. For example, if
        ``root == "my_dir"``, ``filename == "CONTCAR"`` and we have a single
        structure file in ``my_dir/test/CONTCAR``, then the resulting
        structures will be ``{"my_dir/test/CONTCAR": struct}``. Similarly, if
        ``filename == "CONTCAR*", then all files of the form ``CONTCAR*`` will
        be found and used. The directory structure produced will be something
        like

        .. code-block:: python

            {
                "my_dir/test/CONTCAR1": struct1,
                "my_dir/test/CONTCAR2": struct2,
                ...
            }

        Parameters
        ----------
        root : str
            The directory in which to begin the search.
        filename : str, optional
            The files to search for. Uses ``rglob`` to recursively find any
            files matching ``filename`` within the provided directory.
        cleanup_paths : bool, optional
            If True, runs :class:`cleanup_paths()` after initializing the
            Database.

        Returns
        -------
        Database
        """

        structures = {
            str(Path(path.parent) / path.stem): Structure.from_file(
                path
            ).get_primitive_structure()
            for path in Path(root).rglob(filename)
        }

        # Check for any duplicate names
        names = list(structures.keys())
        names = [Path(name).name for name in names]
        if len(list(set(names))) < len(names):
            new_structures = dict()
            for key, value in structures.items():
                new_structures[str(Path(key).parent)] = value
            structures = new_structures

        metadata = {key: dict() for key in structures.keys()}

        kls = cls(structures=structures, metadata=metadata, supercells=dict())
        if cleanup_paths:
            kls.cleanup_paths()
        return kls

    @classmethod
    def from_materials_project(
        cls,
        query,
        query_type="mpids",
        api_key=None,
        metadata_keys=[
            "created_at",
            "blessed_tasks",
            "pseudo_potential",
            "spacegroup",
            "_id",
            "structure",
            "icsd_ids",
            "e_above_hull",
            "formation_energy_per_atom",
            "band_gap",
            "diel",
        ],
        verbose=True,
    ):
        """Constructs the :class:`.Database` object by pulling structures and
        metadata directly from the Materials Project. The following query types
        are allowed:

        * ``query_type == mpids``, the ``query`` argument is simply a list of
          Materials Project IDs.
        * ``query_type == patterns``, the ``query`` argument is a list
          of patterns, e.g. ``[Ti-O, Ti-O-*, Cu-O-*]``.
        * ``query_type == mp_query``, then the ``query`` argument is
          just a dict that is passed directly to the ``mpr.query`` method,
          where ``mpr`` is the MPRester object.

        Parameters
        ----------
        query : list or dict
            The query itself, the form of which is determined by
            ``query_type``.
        query_type : str, optional
            There are three different types of allowed queries: ``mpids``,
            ``patterns``, or ``mp_query``. See above for the allowed query
            types.
        api_key : str, optional
            Materials Project API key. If None (not provided), looks for the
            environment variable ``PMG_API_KEY``.
        metadata_keys : list, optional
            The Materials Project metadata contains a huge amount of
            information. If not ``None``, these are the only keys that are kept
            in the pulled metadata object.
        verbose : bool, optional
            If True, uses the :class:`tqdm` progress bar. Otherwise will
            silence it.

        Returns
        -------
        Database

        Examples
        --------
        Construct a :class:`.Database` via directly pulling certain materials
        by MPID.

        .. code-block:: python

            database = Database.from_materials_project(["mp-390", "mvc-1115"])

        Construct a :class:`.Database` via pulling all materials consistent
        with certain patterns. For example, to pull all binary and ternary
        titanium oxide compounds:

        .. code-block:: python

            database = Database.from_materials_project(
                ["Ti-O", "Ti-O-*"], query_type="patterns"
            )
        """

        api_key = _get_api_key(api_key)

        # Nothing to do here
        if query_type == "mpids":
            mpids = query

        # Convert the patterns into a list of mpids
        elif query_type == "patterns":
            mpids = []
            with MPRester(api_key) as mpr:
                for pattern in query:
                    data = mpr.get_data(pattern, data_type="vasp")
                    mpids.extend([xx["material_id"] for xx in data])

        # Convert the raw query itself to a list of mpids
        elif query_type == "mp_query":
            with MPRester(api_key) as mpr:
                materials_list = mpr.query(**query)
            mpids = [xx["material_id"] for xx in materials_list]

        # Otherwise we error and terminate
        else:
            raise ValueError(f"Unknown query {query_type}")

        # Get all of the data as a dictionary
        data = _from_mpids_list(
            mpids,
            api_key,
            metadata_keys,
            verbose=verbose,
        )

        return cls(
            structures=data["structures"],
            metadata=data["metadata"],
            supercells=dict(),
        )

    def initialize_supercells(self, supercell_cutoff=9.0):
        """Initializes the supercells from the structures pulled from the
        Materials project.

        Parameters
        ----------
        supercell_cutoff : float, optional
            Parameter used for constructing the supercells. Default is 9
            Angstroms. TODO: this needs clearer documentation.
        """

        for key, prim in self._structures.items():
            supercell = pymatgen_utils.make_supercell(prim, supercell_cutoff)
            self._supercells[key] = supercell.get_sorted_structure()
        self._supercell_cutoff = supercell_cutoff
        self._supercells_initialized = True

    def initialize_inequivalent_sites(self):
        """Iterates through the structures and updates the metadata with
        keys corresponding to the inequivalent sites in the structure. This
        also tracks the atom types corresponding to the inequivalent sites and
        their multiplicities in the structure."""

        for key, prim in self._structures.items():
            info = pymatgen_utils.get_inequivalent_site_info(prim)
            self._metadata[key]["primitive"] = info

        # Do the same for the supercells
        for key, supercell in self._supercells.items():
            info = pymatgen_utils.get_inequivalent_site_info(supercell)
            self._metadata[key]["supercell"] = info

        self._inequivalent_sites_initialized = True

    @property
    def structures(self):
        """A dictionary of :class:`pymatgen.core.structure.Structure` objects.
        Contains the primitive structures. The keys are the IDs of the
        structure. In the case of data pulled from the Materials Project, these
        are the MPIDs. Otherwise, they are simply strings encoding some
        information about the origins of the structures.

        Returns
        -------
        dict
        """

        return self._structures

    @property
    def supercells(self):
        """A dictionary of :class:`pymatgen.core.structure.Structure` objects
        containing supercells and the same keys as ``structures``.

        Returns
        -------
        dict
        """

        if not self._supercells_initialized:
            warn("Run initialize_supercells(); supercells is probably empty")

        return self._supercells

    @property
    def metadata(self):
        """A dictionary of metadata information about the structure. This
        information is filled by the Materials Project when data is pulled,
        and might be empty if there is no associated metadata (in cases of
        using data already on disk, for example).

        Returns
        -------
        dict
        """

        return self._metadata

    @property
    def database_status(self):
        """A dictionary containing the current status of the database.
        Basically everything except the structures, metadata and supercells.

        Returns
        -------
        dict
        """

        return {
            key: value
            for key, value in self.as_dict().items()
            if key not in ["structures", "metadata", "supercells"]
        }

    def __init__(
        self,
        structures,
        metadata=dict(),
        supercells=dict(),
        supercell_cutoff=None,
        inequivalent_sites_initialized=False,
        supercells_initialized=False,
    ):
        """Initializer for the Database class. Note it is recommended to use
        the classmethods to initialize this object."""

        self._structures = structures
        self._metadata = metadata
        self._supercells = supercells
        self._supercell_cutoff = supercell_cutoff
        self._inequivalent_sites_initialized = inequivalent_sites_initialized
        self._supercells_initialized = supercells_initialized

    def _setup_preliminary_attributes(self):
        """Initializes supercells and inequivalent site info in the metadata
        if they're not already."""

        if not self._supercells_initialized:
            warn(
                "Initializing supercells with supercell_cutoff=9.0. "
                "Run initialize_supercells(supercell_cutoff=...) to set "
                "this cutoff manually."
            )
            self.initialize_supercells(9.0)

        if not self._inequivalent_sites_initialized:
            self.initialize_inequivalent_sites()

    @staticmethod
    def _get_site_indexes_matching_atom(info, species):
        """For an info dictionary of the form
        {
            "sites": inequivalent_sites,
            "species": species,
            "multiplicities": multiplicities
        }
        gets the inequivalent indexes matching the provided atom type/species.
        Also returns None if ``species`` is None.

        Parameters
        ----------
        info : dict
        species : str
            The atom type, e.g. "Ti", "O", etc.

        Returns
        -------
        list
        """

        if species is None:
            return None

        return [
            index
            for index, specie in zip(info["sites"], info["species"])
            if specie == species
        ]

    def write(
        self,
        root,
        absorbing_atoms=None,
        options=[],
        pbar=True,
        copy_script=None,
    ):
        """The core method of the :class:`.Database` class. This method will
        write all input files specified in the ``options`` parameter to disk.
        Of particular note is the directory structure, which is always
        consistent regardless of the type of calculation: At the first level is
        the ``key`` (usually an mpid) indexing the material of interest. At the
        next level is the user-specified "name" of the calculation, which are
        the first elements of each tuple in the ``options``. Next are
        the material-specific calculations. For example, in spectroscopy
        calculations, each absorbing atom will have it's own directory at this
        level. Within each of those are the input files for that calculation.
        For example::

            mp-390
                VASP
                    000_Ti
                    SCF
                FEFF-XANES
                    000_Ti
                        # ... (input files)
            mvc-11115
                # ...

        .. note::

            At the end of writing the files, a ``writer_metadata.json`` file
            will be saved along with the directories containing the input
            files. This metadata file contains all information about the
            parameters used to construct the input files, including those
            passed as arguments to this method.

        Parameters
        ----------
        root : os.PathLike
            The target root directory to save all of the input files.
        absorbing_atoms : str or list, optional
            The absorbing atom type symbol(s), e.g. ``"Ti"``. Note that if None,
            any calculations in which the absorbing atom is required (e.g. all
            spectroscopy) will be skipped. Only calculations that do not
            require absorbing atoms to be specified (e.g. neutral potential
            VASP electronic structure self-consistent procedure) will be
            performed.
        options : list, optional
            A list of :class:`lightshow.parameters._base._BaseParameters`
            objects or derived instances. The choice of options not only
            specifies which calculations to setup, each of the options also
            contains the complete set of parameters necessary to characterize
            each individual set of input files (for e.g. FEFF, VASP, etc.).
        pbar : bool, optional
            If True, enables the :class:`tqdm` progress bar.
        copy_script : os.PathLike
            If not ``None``, will copy the script in the provided path to each
            of the input file locations.
        """

        self._setup_preliminary_attributes()

        root = str(Path(root).resolve())  # Get absolute path

        if not isinstance(absorbing_atoms, list):
            absorbing_atoms = [absorbing_atoms]

        writer_metadata = {
            "locals": {
                key: value
                for key, value in locals().items()
                if key not in ["self", "options"]
            },
            "created_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "options": [option.as_dict() for option in options],
            **self.database_status,
        }

        root = Path(root)
        root.mkdir(exist_ok=True, parents=True)
        writer_metadata_path = root / Path("writer_metadata.json")

        for key, supercell in tqdm(self._supercells.items(), disable=not pbar):

            for absorbing_atom in absorbing_atoms:

                primitive_info = self._metadata[key]["primitive"]
                supercell_info = self._metadata[key]["supercell"]

                # If inequiv is None, that means that the absorbing_atom was not
                # specified (absorbing_atom is None)
                inequiv = self._get_site_indexes_matching_atom(
                    primitive_info, absorbing_atom
                )
                inequiv_sc = self._get_site_indexes_matching_atom(
                    supercell_info, absorbing_atom
                )

                if inequiv_sc is not None and inequiv is not None:
                    index_mapping = {k: v for k, v in zip(inequiv, inequiv_sc)}
                else:
                    index_mapping = None

                # If the for VASP and XSpectra calculations, use supercell;
                # otherwise, use unit cell structure
                # test if band_gap and diel in the self._metadata[key].keys()
                # if yes, read the bandgap and diel for OCEAN
                # if no, ignore them
                kwargs = {
                    "structure_sc": supercell,
                    "structure_uc": self._structures[key],
                    "sites": inequiv,
                    "index_mapping": index_mapping,
                }
                if key in self._metadata.keys():
                    if (
                        "band_gap" in self._metadata[key].keys()
                        and "diel" in self._metadata[key].keys()
                    ):
                        kwargs["bandgap"] = self._metadata[key]["band_gap"]
                        kwargs["diel"] = self._metadata[key]["diel"]

                # Write the files that we can
                for option in options:
                    path = root / Path(key) / Path(option.name)
                    status = option.write(path, **kwargs)
                    if not status["pass"]:
                        d = {"name": key, **status["errors"]}
                        warn(f"error: {key}+{option.name}: {d}")
                    if copy_script is not None and "paths" in status.keys():
                        for p in status["paths"]:
                            copy2(copy_script, p)

        # Save a metadata file (not a serialized version of this class) to
        # disk along with the input files
        with open(writer_metadata_path, "w") as outfile:
            json.dump(writer_metadata, outfile, indent=4, sort_keys=True)



from pathlib import Path
import pytest
import sys

from lightshow import VASPParameters
from lightshow.defaults import VASP_INCAR_DEFAULT_COREHOLE_POTENTIAL

# Helper testing files
sys.path.append(str(Path(__file__).parent.resolve() / Path("helpers")))
from geometry import consistency_check  # noqa


@pytest.mark.parametrize(
    "mpid",
    ["mp-390", "mvc-11115"],
)
def test_write(mpid, dummy_potcar_file_directory, database_from_file, tmp_path):

    # Load it all in
    dat = database_from_file
    dat.initialize_supercells(9.0)
    dat.initialize_inequivalent_sites()
    dat._supercells = {mpid: dat.supercells[mpid]}
    dat._structures = {mpid: dat.structures[mpid]}
    dat._metadata = {mpid: dat.metadata[mpid]}

    target = Path(tmp_path) / Path("iama") / Path("destination")
    target.mkdir(exist_ok=True, parents=True)

    vasp_params_corehole = VASPParameters(
        incar=VASP_INCAR_DEFAULT_COREHOLE_POTENTIAL,
        potcar_directory="this/dir/does/not/exist",
        force_spin_unpolarized=False,
    )

    dat.write(
        target,
        absorbing_atoms=["Ti", "O"],
        options=[vasp_params_corehole],
        pbar=False,
    )



from abc import ABC, abstractmethod

from monty.json import MSONable


class _BaseNbandsMethod(ABC):
    @abstractmethod
    def __call__(self, structure):
        """Base abstraction for the various ways to get the number of bands
        used in a variety of alculations.

        Parameters
        ----------
        structure : pymatgen.core.structure.Structure

        Returns
        -------
        int
        """

        ...


class Constant(MSONable, _BaseNbandsMethod):
    def __init__(self, nbands):
        self._nbands = nbands

    def __call__(self, structure):
        return int(self._nbands)


class UnitCellVolumeEstimate(MSONable, _BaseNbandsMethod):
    """Return a guess at the number of conduction bands that a given
    unit-cell volume needs to cover a given energy range.

    Parameters
    ----------
    e_range : float
        The energy range in eV.
    """

    def __init__(self, e_range=40.0):
        self._e_range = e_range

    def __call__(self, structure):
        return int(
            round(
                0.256
                * structure.lattice.volume
                * ((self._e_range / 13.6056980659) ** (3.0 / 2.0))
            )
        )


# class LightshowHEG(MSONable, _BaseNbandsMethod):

#     def __init__(self, e_range=40.0):
#         self._e_range = e_range

#     def __call__(self, structure, total_valence_electrons):
#         """Gets an approximation for the total number of bands to use in the
#         calculation based on the free electron gas model, such that a range of
#         ``eRange`` is covered in the spectrum. The method assumes all
#         electrons
#         are free electrons, including those in the valence bands. The total
#         number of electrons will then be calculated as
#         :math:`N = N_\\mathrm{v} + N_\\mathrm{c}`, where :math:`N_\\mathrm{v}`
#         is the total number of valence electrons and
#         :math:`N_\\mathrm{c}` is the number of conduction band electrons above
#         the Fermi energy as specified by the provided energy range. The final
#         expression (in atomic units) is

#         .. math::

#             E_\\mathrm{range}/c_1 = c_2 / (V/N_vc * 3/ (4 pi))^2/3
#             - c_2 / (V/N_v * 3 / (4 pi))^2/3 N_v

#         where :math:`c_1 = 27.2114` and :math:`c_2 = 1.841`,
#         is summed over all valence electrons obtained from POTCAR.

#         Parameters
#         ----------
#         structure : pymatgen.core.structure.Structure
#         eRange : float
#             The approximate energy range desired in the spectrum, starting at
#             the Fermi energy.

#         Returns
#         -------
#         int
#             Approximation for the number of bands to use in computing the
#             XANES
#             spectrum as computed using the (3D) homogeneous electron gas
#             approximation.
#         """

#         ANG2BOHR = 1.88873
#         volume_bohr = structure.volume * ANG2BOHR**3
#         Nv = self.get_total_valence_electrons(structure)

#         tmp1 = self._e_range / 27.2114 + 1.841 / (
#             volume_bohr / total_valence_electrons * 3.0 / 4.0 / np.pi
#         ) ** (2 / 3)
#         tmp2 = (1.841 / tmp1) ** (2 / 3) / 3.0 * 4.0 * np.pi
#         Nvc = volume_bohr / tmp2 * 0.65
#         return int(np.ceil(Nvc))



from pathlib import Path
import json
import bz2
import base64
import shutil
from warnings import warn

from monty.json import MSONable
from pymatgen.io.pwscf import PWInput

from lightshow.parameters._base import _BaseParameters
from lightshow.common.kpoints import GenericEstimatorKpoints
from lightshow import (
    _get_CHPSP_DIRECTORY_from_environ,
    _get_PSP_DIRECTORY_from_environ,
)

XSPECTRA_DEFAULT_CARDS = {
    "QE": {
        "control": {"restart_mode": "from_scratch"},
        "electrons": {"conv_thr": 1e-08, "mixing_beta": 0.4},
        "system": {
            "degauss": 0.002,
            "ecutrho": 320,
            "ecutwfc": 40,
            "nspin": 1,
            "occupations": "smearing",
            "smearing": "gauss",
        },
    },
    "XS": {
        "cut_occ": {"cut_desmooth": 0.3},
        "input_xspectra": {
            "outdir": "../",
            "prefix": "pwscf",
            "xcheck_conv": 200,
            "xerror": 0.01,  #
            "xniter": 5000,
            "xcoordcrys": ".false.",
        },
        "kpts": {"kpts": "2 2 2", "shift": "0 0 0"},
        "plot": {
            "cut_occ_states": ".true.",
            "terminator": ".true.",
            "xemax": 70,
            "xemin": -15.0,
            "xnepoint": 400,
        },
    },
}


class XSpectraParameters(MSONable, _BaseParameters):
    """A one-stop-shop for all the different ways to modify input parameters
    for an XSpectra calculation.

    Parameters
    ----------
    cards : dict
        A dictionary of of the cards to be control the parameters in the
        OCEAN calculations. For example, one might wish to use something like

        .. code-block:: python

           cards = {
                    "QE": {
                        "control": {
                            "restart_mode": "from_scratch",
                        },
                        "electrons": {"conv_thr": 1e-08, "mixing_beta": 0.4},
                        "system": {
                            "degauss": 0.002,
                            "ecutrho": 320,
                            "ecutwfc": 40,
                            "nspin": 1,
                            "occupations": "smearing",
                            "smearing": "gauss",
                        },
                    },
                    "XS": {
                        "cut_occ": {"cut_desmooth": 0.3},
                        "input_xspectra": {
                            "outdir": "../",
                            "prefix": "pwscf",
                            "xcheck_conv": 200,
                            "xerror": 0.01,  #
                            "xniter": 5000,
                            "xcoordcrys": ".false.",
                       },
                        "kpts": {"kpts": "2 2 2", "shift": "0 0 0"},
                        "plot": {
                            "cut_occ_states": ".true.",
                            "terminator": ".true.",
                            "xemax": 70,
                            "xemin": -15.0,
                            "xnepoint": 400,
                        },
                    },
            }

    kpoints : lightshow.common.kpoints._BaseKpointsMethod
        The method for constructing he kpoints file from the structure. Should
        be a class with a ``__call__`` method defined. This method should take
        the structure as input and return a tuple corresponding to the kpoints
        density along each axis.
    psp_directory : os.PathLike, optional
        The location in which the neutral potential files for absorption atoms
        are stored. In the psp_directory, a cutoff table should also be
        provided. The cutoff table should have similar structure as the one for
        SSSP database. The name of the cutoff table needs to be given in
        cards['XS']['psp_cutoff_table']. If None, checks the  environment for
        ``XS_PSP_DIRECTORY``.
    chpsp_directory : os.PathLike, optional
        The location in which the core-hole potential files for absorption atoms
        are stored. Each element should have two files, e.g. "Ti.fch.upf" and
        "Core_Ti.wfc". "Ti.fch.upf" is the core-hole pesuodo potetial file and
        "Core_Ti.wfc" is the core electron wavefunction. The naming of the
        pseudopotentials and core electron wavefunction should follow the exact
        specific structure. If None, checks the environment for
        ``XS_CHPSP_DIRECTORY``.
    psp_cutoff_table  : str
        The name of the cutoff table to loop around the pseudo potentials, the
        format should be the same as the one for SSSP database. The cutoff
        table shoyld be placed under the psp_directory.
    psp_json : str
        The name of the compact pseudo potential database in a single json
        file, which can be generated by the packPsps method. If not None, the
        json file should be placed under the psp_directory.
    """

    @property
    def name(self):
        return self._name

    @property
    def cards(self):
        return self._cards

    def __init__(
        self,
        cards=XSPECTRA_DEFAULT_CARDS,
        kpoints=GenericEstimatorKpoints(cutoff=16.0, max_radii=50.0),
        chpsp_directory=None,
        psp_directory=None,
        psp_cutoff_table=None,
        psp_json=None,
        defaultConvPerAtom=1e-10,
        edge="K",
        name="XSpectra",
    ):
        # Default cards
        self._cards = cards

        # chpsp information
        if chpsp_directory is None:
            chpsp_directory = _get_CHPSP_DIRECTORY_from_environ()
        if chpsp_directory is None:
            warn(
                "chpsp_directory not provided, and XS_CHPSP_DIRECTORY not in "
                "the current environment variables. core-hole pseudo "
                "potential files will not be written."
            )
        self._chpsp_directory = chpsp_directory
        # psp information
        self._psp_cutoff_table = psp_cutoff_table
        self._psp_json = psp_json

        if psp_directory is None:
            psp_directory = _get_PSP_DIRECTORY_from_environ()

        self._psp_directory = psp_directory

        if psp_directory is None or self._psp_cutoff_table is None:
            warn(
                "psp_directory not provided XS_PSP_DIRECTORY not in the "
                "current environment variables OR psp_cutoff_table not "
                "provided. neutral pseudo potential files will not be written."
            )

        # Method for determining the kmesh
        self._kpoints = kpoints
        self._defaultConvPerAtom = defaultConvPerAtom
        self._edge = edge
        self._name = name

    def packPsps(self, pspJsonOut):
        """This method packs all the pseudo potentials the self._psp_directory
        into a single json file, whose name is given by ``pspJsonOut``. By using
        the condensed json file, the performance for writing pseudo potentials
        files is better than copying them from self._psp_directory to the
        working directory.

        Parameters
        ----------
        pspJsonOut : str
            Name of the output json file. Extension should be included.
        """
        cutofftable = Path(self._psp_directory) / Path(self._psp_cutoff_table)
        with open(cutofftable, "r") as f:
            inJSON = json.load(f)

        outJSON = dict()

        for element in inJSON:
            pspFile = Path(self._psp_directory) / Path(
                inJSON[element]["filename"]
            )
            with open(pspFile, "r") as f:
                pspString = f.read()

            outJSON[inJSON[element]["filename"]] = base64.b64encode(
                bz2.compress(pspString.encode("utf-8"), 9)
            ).decode("utf-8")

        with open(pspJsonOut, "w") as f:
            f.write(json.dumps(outJSON))

    def _unpackPsps(
        self,
        ecutwfc,
        ecutrho,
        symbols,
        folder,
    ):
        """Used to generate the pseudo potential files from the compact json
        file.

        Parameters
        ----------
        ecutwfc : float
            Energy cutoff for wave function
        ecutrho : float
            Energy cutoff for charge density
        symbols : list
            List of symbols for the element in the strcuture
        folder : str
            Name of the folder where the pseudo potential files will be
            recovered

        Returns
        -------
        psp : dict
            A dictionary with key as the name of element and value as the name
            of the pseudo potential file name.
        ecutwfc : float
            Energy cutoff for wave function
        ecutrho : float
            Energy cutoff for charge density
        """
        psp = {}
        sssp_fn = Path(self._psp_directoy) / Path(self._psp_cutoff_table)
        with open(sssp_fn, "r") as pspDatabaseFile:
            pspDatabase = json.load(pspDatabaseFile)
        minSymbols = set(symbols)
        for symbol in minSymbols:
            psp[symbol] = pspDatabase[symbol]["filename"]
            if ecutwfc < pspDatabase[symbol]["cutoff"]:
                ecutwfc = pspDatabase[symbol]["cutoff"]
            if ecutrho < pspDatabase[symbol]["rho_cutoff"]:
                ecutrho = pspDatabase[symbol]["rho_cutoff"]

        sssp_fn = Path(self._psp_directory) / Path(self._psp_json)
        with open(sssp_fn, "r") as p:
            pspJSON = json.load(p)
        for symbol in minSymbols:
            fileName = psp[symbol]
            pspString = bz2.decompress(base64.b64decode(pspJSON[fileName]))
            # print("Expected hash:  " + pspDatabase[symbol]["md5"])
            # print("Resultant hash: " + hashlib.md5(pspString).hexdigest())
            with open(folder / fileName, "w") as f:
                f.write(pspString.decode("utf-8"))

        return psp, ecutwfc, ecutrho

    def _write_xspectra_in(self, mode, iabs, dirs, xkvec):
        """construct input file for XSpectra calculation

        Parameters
        ----------
        mode : str
            "dipole" or "quadrupole"
        iabs : int
            The index of the absorbing element in scf calculation
        dirs : list
            Description of the polarization direction, e.g. [1,0,0]
            corresponds to the x direction
        xkvec : list
            Description of the k vectors for quadrupole calculation

        Returns
        -------
        string of the XSpectra input file
        """
        XSparams = self._cards["XS"]
        element = XSparams["element"]
        inp = [
            "&input_xspectra",
            "    calculation = 'xanes_%s'" % mode,
            "    edge = '" + XSparams["input_xspectra"]["edge"] + "'",
            "    prefix = 'pwscf'",
            "    outdir = '../'",
            "    xniter = " + str(XSparams["input_xspectra"]["xniter"]),
            "    xiabs = %d" % iabs,
            "    xerror = " + str(XSparams["input_xspectra"]["xerror"]),
            "    xcoordcrys = " + XSparams["input_xspectra"]["xcoordcrys"],
            "    xcheck_conv = "
            + str(XSparams["input_xspectra"]["xcheck_conv"]),
            "    xepsilon(1) = %d" % dirs[0],
            "    xepsilon(2) = %d" % dirs[1],
            "    xepsilon(3) = %d" % dirs[2],
        ]

        if mode == "quadrupole":
            inp += [
                "    xkvec(1) = %.10f" % xkvec[0],
                "    xkvec(2) = %.10f" % xkvec[1],
                "    xkvec(3) = %.10f" % xkvec[2],
            ]

        inp += [
            "/",
            "&plot",
            "    xnepoint = " + str(XSparams["plot"]["xnepoint"]),
            "    xemin = " + str(XSparams["plot"]["xemin"]),
            "    xemax = " + str(XSparams["plot"]["xemax"]),
            "    terminator = " + XSparams["plot"]["terminator"],
            "    cut_occ_states = " + XSparams["plot"]["cut_occ_states"],
            # use very small smearing value: 0.01 eV
            "    gamma_mode = 'constant'",
            "    xgamma = 0.01 ",
            "/",
        ]

        inp += [
            "&pseudos",
            f"    filecore = '../../Core_{element}.wfc'",
            "/",
            "&cut_occ",
            "    cut_desmooth = " + str(XSparams["cut_occ"]["cut_desmooth"]),
            "/",
            XSparams["kpts"]["kpts"] + " " + XSparams["kpts"]["shift"],
        ]
        return "\n".join(inp) + "\n"

    def write(self, target_directory, **kwargs):
        """Writes the input files for the provided structure and sites.

        Parameters
        ----------
        target_directory : os.PathLike
            The target directory to which to save the FEFF input files.
        **kwargs
            Must contain the ``structure_sc`` key (the
            :class:`pymatgen.core.structure.Structure` of interest), the
            ``sites`` key (a list of int, where each int corresponds to the
            site index in the supercell structure) and the ``index_mapping``
            key (a dictionary mapping the index between unit cell and supercell)

        Returns
        -------
        dict
            A dictionary containing the status and errors key. In the case of
            XSpectra, there are no possible errors at this stage other than
            critical ones that would cause program termination, so the returned
            object is always ``{"pass": True, "errors": dict()}``.
        """

        structure = kwargs["structure_sc"]
        sites = kwargs["sites"]
        index_mapping = kwargs["index_mapping"]

        target_directory = Path(target_directory)
        target_directory.mkdir(exist_ok=True, parents=True)
        symbols = [spec.symbol for spec in structure.species]
        # Obtain absorbing atom
        species = [
            structure[index_mapping[site]].specie.symbol for site in sites
        ]
        element = species[0]
        self._cards["XS"]["element"] = element
        self._cards["XS"]["input_xspectra"]["edge"] = self._edge
        symTarg = element
        # Estimate number of kpoints
        if len(structure.get_primitive_structure()) != len(structure):
            # use Gamma point for ground state calculations (es.in and gs.in)
            kpoints_scf = [1, 1, 1]
        else:
            kpoints_scf = self._kpoints(structure)

        kpoints_xas = self._kpoints(structure)

        self._cards["XS"]["kpts"][
            "kpts"
        ] = f"{kpoints_xas[0]} {kpoints_xas[1]} {kpoints_xas[2]}"
        # Determine the SCF? convergence threshold
        self._cards["QE"]["electrons"][
            "conv_thr"
        ] = self._defaultConvPerAtom * len(structure)
        # Get the psp data ready for the GS calculations; similar to SCF
        # (neutral) calculations in VASP
        # need to treat three different cases here

        ecutwfc = self._cards["QE"]["system"]["ecutwfc"]
        ecutrho = self._cards["QE"]["system"]["ecutrho"]

        if self._psp_directory is not None:
            try:
                psp_dict = json.load(
                    open(
                        Path(self._psp_directory) / Path(self._psp_cutoff_table)
                    )
                )
                psp = dict()
                for symbol in symbols:
                    psp_filename = psp_dict[symbol]["filename"]
                    psp[symbol] = psp_filename
                    shutil.copyfile(
                        Path(self._psp_directory) / Path(psp_filename),
                        target_directory / psp_filename,
                    )
                    if psp_dict[symbol]["cutoff_wfc"] > ecutwfc:
                        ecutwfc = psp_dict[symbol]["cutoff_wfc"]
                    if psp_dict[symbol]["cutoff_rho"] > ecutrho:
                        ecutrho = psp_dict[symbol]["cutoff_rho"]
            except FileNotFoundError:
                try:
                    psp, ecutwfc, ecutrho = self._unpackPsps(
                        ecutwfc,
                        ecutrho,
                        symbols,
                        target_directory,
                    )
                except FileNotFoundError:
                    warn(
                        "Some pseudo potential files are not present in "
                        f"f{self._psp_directory}"
                    )
                    self._psp_directory = None

        if self._psp_directory is None:
            psp = {symbol: symbol + ".upf" for symbol in symbols}

        self._cards["QE"]["system"]["ecutwfc"] = ecutwfc
        self._cards["QE"]["system"]["ecutrho"] = ecutrho
        self._cards["QE"]["control"]["pseudo_dir"] = "../"

        path = target_directory / "GS"
        path.mkdir(exist_ok=True, parents=True)
        gs_in = PWInput(
            structure,
            pseudo=psp,
            control=self._cards["QE"]["control"],
            system=self._cards["QE"]["system"],
            electrons=self._cards["QE"]["electrons"],
            kpoints_grid=kpoints_scf,
        )
        gs_in.write_file(path / "gs.in")

        psp[f"{element}+"] = f"{element}.fch.upf"  # psp2[i]
        # copy core-hole potential and core wfn to target folder
        if self._chpsp_directory is not None:
            try:
                shutil.copyfile(
                    self._chpsp_directory + f"/{element}.fch.upf",
                    target_directory / Path("{element}.fch.upf"),
                )
                shutil.copyfile(
                    self._chpsp_directory + f"/Core_{element}.wfc",
                    target_directory / Path("Core_{element}.wfc"),
                )
            except FileNotFoundError:
                warn(
                    f"{element}.fch.upf or Core_{element}.wfc not found "
                    f"in {self._chpsp_directory}"
                )

        # Determine iabs
        for i, j in enumerate(sorted(psp.keys())):
            if j == symTarg + "+":
                iabs = i + 1
        for site, specie in zip(sites, species):
            path = target_directory / Path(f"{site:03}_{specie}")
            path.mkdir(exist_ok=True, parents=True)

            structure[index_mapping[site]] = element + "+"
            self._cards["QE"]["control"]["pseudo_dir"] = "../"
            es_in = PWInput(
                structure,
                pseudo=psp,
                control=self._cards["QE"]["control"],
                system=self._cards["QE"]["system"],
                electrons=self._cards["QE"]["electrons"],
                kpoints_grid=kpoints_scf,
            )
            es_in.write_file(path / "es.in")
            structure[index_mapping[site]] = element

            # Deal with the dipole case only
            # notice I put the photonSymm in the folder, which is created by
            # John
            photons = list()
            photons.append({"dipole": [1, 0, 0, 1]})
            photons.append({"dipole": [0, 1, 0, 1]})
            photons.append({"dipole": [0, 0, 1, 1]})

            totalweight = 0
            for photon in photons:
                totalweight += photon["dipole"][3]

            photonCount = 0
            for photon in photons:
                photonCount += 1
                dir1 = photon["dipole"][0:3]
                dir2 = dir1
                weight = photon["dipole"][3] / totalweight
                mode = "dipole"

                xanesfolder = path / f"{mode}{photonCount}"
                xanesfolder.mkdir(parents=True, exist_ok=True)
                with open(xanesfolder / "xanes.in", "w") as f:
                    f.write(
                        self._write_xspectra_in(
                            mode,
                            iabs,
                            dir1,
                            dir2,
                        )
                    )

                with open(xanesfolder / "weight.txt", "w") as f:
                    f.write(str(weight) + "\n")

        return {"pass": True, "errors": dict()}



from abc import ABC, abstractmethod


class _BaseParameters(ABC):
    @abstractmethod
    def write(self, target_directory, kwargs):
        ...

    @property
    def name(self):
        return self._name
